from typing import Union

import numpy as np
import pandas as pd
from astropy.time import Time
from bilby.core.prior import PriorDict
from bilby.gw.detector import InterferometerList
from torchvision.transforms import Compose

from dingo.core.samplers import Sampler, GNPESampler
from dingo.core.transforms import GetItem, RenameKey
from dingo.gw.domains import build_domain
from dingo.gw.gwutils import get_window_factor
from dingo.gw.result import Result
from dingo.gw.transforms import (
    WhitenAndScaleStrain,
    RepackageStrainsAndASDS,
    ToTorch,
    SelectStandardizeRepackageParameters,
    GNPECoalescenceTimes,
    TimeShiftStrain,
    GNPEBase,
    PostCorrectGeocentTime,
    CopyToExtrinsicParameters,
    GetDetectorTimes,
)


class GWSamplerMixin(object):
    """
    Mixin class designed to add gravitational wave functionality to Sampler classes:
        * builder for data domain
        * correction for fixed detector locations during training (t_ref)
    """

    def __init__(self, **kwargs):
        """
        Parameters
        ----------
        kwargs
            Keyword arguments that are forwarded to the superclass.
        """
        super().__init__(**kwargs)
        self.t_ref = self.base_model_metadata["train_settings"]["data"]["ref_time"]
        self._pesummary_package = "gw"
        self._result_class = Result

    def _build_domain(self):
        """
        Construct the domain object based on model metadata. Includes the window factor
        needed for whitening data.

        Called by __init__() immediately after _build_prior().
        """
        self.domain = build_domain(
            self.base_model_metadata["dataset_settings"]["domain"]
        )

        data_settings = self.base_model_metadata["train_settings"]["data"]
        if "domain_update" in data_settings:
            self.domain.update(data_settings["domain_update"])

        self.domain.window_factor = get_window_factor(data_settings["window"])

    def _correct_reference_time(
        self, samples: Union[dict, pd.DataFrame], inverse: bool = False
    ):
        """
        Correct the sky position of an event based on the reference time of the model.
        This is necessary since the model was trained with with fixed detector (reference)
        positions. This transforms the right ascension based on the e difference between
        the time of the event and t_ref.

        The correction is only applied if the event time can be found in self.metadata[
        'event'].

        This method modifies the samples in place.

        Parameters
        ----------
        samples : dict or pd.DataFrame
        inverse : bool, default True
            Whether to apply instead the inverse transformation. This is used prior to
            calculating the log_prob.
        """
        # FIXME: Update this method to make sure that for models that do not include sky
        #  position, it does not do anything.
        if self.event_metadata is not None:
            t_event = self.event_metadata.get("time_event")
            if t_event is not None and t_event != self.t_ref:
                ra = samples["ra"]
                time_reference = Time(self.t_ref, format="gps", scale="utc")
                time_event = Time(t_event, format="gps", scale="utc")
                longitude_event = time_event.sidereal_time("apparent", "greenwich")
                longitude_reference = time_reference.sidereal_time(
                    "apparent", "greenwich"
                )
                delta_longitude = longitude_event - longitude_reference
                ra_correction = delta_longitude.rad
                if not inverse:
                    samples["ra"] = (ra + ra_correction) % (2 * np.pi)
                else:
                    samples["ra"] = (ra - ra_correction) % (2 * np.pi)

    def _post_process(self, samples: Union[dict, pd.DataFrame], inverse: bool = False):
        """
        Post processing of parameter samples.
        * Correct the sky position for a potentially fixed reference time.
          (see self._correct_reference_time)
        * Potentially sample a synthetic phase. (see self._sample_synthetic_phase)

        This method modifies the samples in place.

        Parameters
        ----------
        samples : dict or pd.DataFrame
        inverse : bool, default True
            Whether to apply instead the inverse transformation. This is used prior to
            calculating the log_prob.
        """
        if not self.unconditional_model:
            self._correct_reference_time(samples, inverse)
        # if not inverse:
        #     self._correct_reference_time(samples, inverse)
        #     # if self.synthetic_phase_kwargs is not None:
        #     #     print(f"Sampling synthetic phase.")
        #     #     t0 = time.time()
        #     #     self._sample_synthetic_phase(samples, inverse)
        #     #     print(f"Done. This took {time.time() - t0:.2f} seconds.")
        #
        # # If inverting, we go in reverse order.
        # else:
        #     # if self.synthetic_phase_kwargs is not None:
        #     #     self._sample_synthetic_phase(samples, inverse)
        #     self._correct_reference_time(samples, inverse)


class GWSampler(GWSamplerMixin, Sampler):
    """
    Sampler for gravitational-wave inference using neural posterior estimation.
    Augments the base class by defining transform_pre and transform_post to prepare
    data for the inference network.

    transform_pre :
        * Whitens strain.
        * Repackages strain data and the inverse ASDs (suitably scaled) into a torch
          tensor.

    transform_post :
        * Extract the desired inference parameters from the network output (
          array-like), de-standardize them, and repackage as a dict.

    Also mixes in GW functionality for building the domain and correcting the reference
    time.

    Allows for conditional and unconditional models, and draws samples from the model
    based on (optional) context data.

    This is intended for use either as a standalone sampler, or as a sampler producing
    initial sample points for a GNPE sampler.
    """

    def _initialize_transforms(self):

        # preprocessing transforms:
        #   * whiten and scale strain (since the inference network expects standardized
        #   data)
        #   * repackage strains and asds from dicts to an array
        #   * convert array to torch tensor on the correct device
        #   * extract only strain/waveform from the sample
        self.transform_pre = Compose(
            [
                WhitenAndScaleStrain(self.domain.noise_std),
                # Use base metadata so that unconditional samplers still know how to
                # transform data, since this transform is used by the GNPE sampler as
                # well.
                RepackageStrainsAndASDS(
                    self.base_model_metadata["train_settings"]["data"]["detectors"],
                    first_index=self.domain.min_idx,
                ),
                ToTorch(device=self.model.device),
                GetItem("waveform"),
            ]
        )

        # postprocessing transforms:
        #   * de-standardize data and extract inference parameters
        self.transform_post = SelectStandardizeRepackageParameters(
            {"inference_parameters": self.inference_parameters},
            self.metadata["train_settings"]["data"]["standardization"],
            inverse=True,
            as_type="dict",
        )


class GWSamplerGNPE(GWSamplerMixin, GNPESampler):
    """
    Gravitational-wave GNPE sampler. It wraps a PosteriorModel and a standard Sampler for
    initialization. The former is used to generate initial samples for Gibbs sampling.

    Compared to the base class, this class implements the required transforms for
    preparing data and parameters for the network. This includes GNPE transforms,
    data processing transforms, and standardization/de-standardization of parameters.

    A GNPE network is conditioned on additional "proxy" context theta^, i.e.,

    p(theta | theta^, d)

    The theta^ depend on theta via a fixed kernel p(theta^ | theta). Combining these
    known distributions, this class uses Gibbs sampling to draw samples from the joint
    distribution,

    p(theta, theta^ | d)

    The advantage of this approach is that we are allowed to perform any transformation of
    d that depends on theta^. In particular, we can use this freedom to simplify the
    data, e.g., by aligning data to have merger times = 0 in each detector. The merger
    times are unknown quantities that must be inferred jointly with all other
    parameters, and GNPE provides a means to do this iteratively. See
    https://arxiv.org/abs/2111.13139 for additional details.

    Gibbs sampling breaks access to the probability density, so this must be recovered
    through other means. One way is to train an unconditional flow to represent p(theta^
    | d) for fixed d based on the samples produced through the GNPE Gibbs sampling.
    Starting from these, a single Gibbs iteration gives theta from the GNPE network,
    along with the probability density in the joint space. This is implemented in
    GNPESampler provided the init_sampler provides proxies directly and num_iterations
    = 1.

    Attributes (beyond those of Sampler)
    ------------------------------------
    init_sampler : Sampler
        Used for providing initial samples for Gibbs sampling.
    num_iterations : int
        Number of Gibbs iterations to perform.
    iteration_tracker : IterationTracker
        **not set up**
    remove_init_outliers : float
        **not set up**
    """

    def _initialize_transforms(self):
        """
        Builds the transforms that are used in the GNPE loop.
        """
        data_settings = self.metadata["train_settings"]["data"]
        ifo_list = InterferometerList(data_settings["detectors"])

        gnpe_time_settings = data_settings.get("gnpe_time_shifts")
        gnpe_chirp_settings = data_settings.get("gnpe_chirp")
        gnpe_phase_settings = data_settings.get("gnpe_phase")
        if (
            not gnpe_time_settings
            and not gnpe_chirp_settings
            and not gnpe_phase_settings
        ):
            raise KeyError(
                "GNPE inference requires network trained for either chirp mass, "
                "coalescence time, or phase GNPE."
            )

        # transforms for gnpe loop, to be applied prior to sampling step:
        #   * reset the sample (e.g., clone non-gnpe transformed waveform)
        #   * blurring detector times to obtain gnpe proxies
        #   * shifting the strain by - gnpe proxies
        #   * repackaging & standardizing proxies to sample['context_parameters']
        #     for conditioning of the inference network
        transform_pre = [RenameKey("data", "waveform")]
        if gnpe_time_settings:
            transform_pre.append(
                GNPECoalescenceTimes(
                    ifo_list,
                    gnpe_time_settings["kernel"],
                    gnpe_time_settings["exact_equiv"],
                    inference=True,
                )
            )
            transform_pre.append(TimeShiftStrain(ifo_list, self.domain))
        transform_pre.append(
            SelectStandardizeRepackageParameters(
                {"context_parameters": data_settings["context_parameters"]},
                data_settings["standardization"],
                device=self.model.device,
            )
        )
        transform_pre.append(RenameKey("waveform", "data"))

        # Extract GNPE information (list of parameters, dict of kernels) from the
        # transforms.
        self.gnpe_parameters = []
        self.gnpe_kernel = PriorDict()
        for transform in transform_pre:
            if isinstance(transform, GNPEBase):
                self.gnpe_parameters += transform.input_parameter_names
                for k, v in transform.kernel.items():
                    self.gnpe_kernel[k] = v
        print("GNPE parameters: ", self.gnpe_parameters)
        print("GNPE kernel: ", self.gnpe_kernel)

        self.transform_pre = Compose(transform_pre)

        # transforms for gnpe loop, to be applied after sampling step:
        #   * de-standardization of parameters
        #   * post correction for geocent time (required for gnpe with exact equivariance)
        #   * computation of detectortimes from parameters (required for next gnpe
        #       iteration)
        self.transform_post = Compose(
            [
                SelectStandardizeRepackageParameters(
                    {"inference_parameters": self.inference_parameters},
                    data_settings["standardization"],
                    inverse=True,
                    as_type="dict",
                ),
                PostCorrectGeocentTime(),
                CopyToExtrinsicParameters(
                    "ra", "dec", "geocent_time", "chirp_mass", "mass_ratio", "phase"
                ),
                GetDetectorTimes(ifo_list, data_settings["ref_time"]),
            ]
        )

    def _kernel_log_prob(self, samples):
        # TODO: Reimplement as a method of GNPEBase.
        if len({"chirp_mass", "mass_ratio", "phase"} & self.gnpe_kernel.keys()) > 0:
            raise NotImplementedError("kernel log_prob only implemented for time gnpe.")
        gnpe_proxies_diff = {
            k: np.array(samples[k] - samples[f"{k}_proxy"])
            for k in self.gnpe_kernel.keys()
        }
        return self.gnpe_kernel.ln_prob(gnpe_proxies_diff, axis=0)



"""
Step 1: Train unconditional nde
Step 2: Set up likelihood and prior
"""
from pathlib import Path

import yaml
from os import rename, makedirs
from os.path import dirname, join, isfile, exists
import argparse

from dingo.core.models import PosteriorModel
from dingo.gw.result import Result
from dingo.gw.inference.gw_samplers import GWSampler
from dingo.gw.importance_sampling.diagnostics import plot_diagnostics


def parse_args():
    parser = argparse.ArgumentParser(
        description="Importance sampling (IS) for dingo models."
    )
    parser.add_argument(
        "--settings",
        type=str,
        required=True,
        help="Path to settings file.",
    )
    parser.add_argument(
        "--outdir",
        type=str,
        default=None,
        help="Output directory for the unconditional nde and IS results.",
    )
    args = parser.parse_args()

    if args.outdir is None:
        args.outdir = dirname(args.settings)

    return args


def main():
    # parse args, load settings, load dingo parameter samples
    args = parse_args()
    with open(args.settings, "r") as fp:
        settings = yaml.safe_load(fp)
    try:
        result = Result(file_name=settings["parameter_samples"])
    except KeyError:
        # except statement for backward compatibility
        result = Result(file_name=settings["nde"]["data"]["parameter_samples"])
    metadata = result.settings
    samples = result.samples
    # for time marginalization, we drop geocent time from the samples
    inference_parameters = metadata["train_settings"]["data"][
        "inference_parameters"
    ].copy()
    time_marginalization_kwargs = settings.get("time_marginalization", None)
    time_marginalization = time_marginalization_kwargs is not None
    phase_marginalization_kwargs = settings.get("phase_marginalization", None)
    phase_marginalization = phase_marginalization_kwargs is not None
    calibration_marginalization_kwargs = settings.get(
        "calibration_marginalization", None
    )
    synthetic_phase_kwargs = settings.get("synthetic_phase", None)
    synthetic_phase = synthetic_phase_kwargs is not None
    # if sum([time_marginalization, phase_marginalization, synthetic_phase]) > 1:
    #    raise NotImplementedError(
    #        "Only one of time_marginalization, phase_marginalization and"
    #        "synthetic_phase can be set to True."
    #    )
    if time_marginalization and "geocent_time" in samples:
        if "geocent_time" in inference_parameters:
            samples.drop("geocent_time", axis=1, inplace=True)
            inference_parameters.remove("geocent_time")
    if phase_marginalization or synthetic_phase:
        if "phase" in inference_parameters:
            samples.drop("phase", axis=1, inplace=True)
            inference_parameters.remove("phase")
    if "nde" in settings:
        settings["nde"]["data"]["inference_parameters"] = inference_parameters
        settings["nde"]["data"]["parameters"] = inference_parameters
        # TODO: train_unconditional_density_estimator should not accept
        #  settings["data"]["parameters"], such that the line above can be removed.

    # Step 1: Build proposal distribution.
    #
    # We use the dingo posterior as our proposal distribution. We need to be able to
    # sample from, and evaluate this distribution. Here, we therefore train an
    # unconditional neural density estimator (nde) to recover the posterior density from
    # dingo samples. This typically required, since one loses the tractable likelihood
    # when using GNPE for dingo. This is not a big deal though, since we can cheaply
    # sample from the dingo posterior, such that one can easily train an unconditional
    # nde to recover the posterior density.

    if "log_prob" not in samples.columns:
        # Use GPS time as name for now.
        event_name = str(result.event_metadata["time_event"])
        nde_name = settings["nde"].get(
            "path", join(args.outdir, f"nde-{event_name}.pt")
        )
        if isfile(nde_name):
            print(f"Loading nde at {nde_name} for event {event_name}.")
            nde = PosteriorModel(
                nde_name,
                device=settings["nde"]["training"]["device"],
                load_training_info=False,
            )
        else:
            print(f"Training new nde for event {event_name}.")
            nde = result.train_unconditional_flow(
                inference_parameters,
                settings["nde"],
                train_dir=args.outdir,
            )
            print(f"Renaming trained nde model to {nde_name}.")
            rename(join(args.outdir, "model_latest.pt"), nde_name)

        # Step 1a: Sample from proposal.
        nde_sampler = GWSampler(model=nde)
        nde_sampler.run_sampler(num_samples=settings["num_samples"])
        result = nde_sampler.to_result()

    # else:
    #     nde_sampler = GWSamplerUnconditional(
    #         result=result,
    #         synthetic_phase_kwargs=synthetic_phase_kwargs,
    #     )
    #
    # # Step 2: Sample from proposal.
    #
    # print(f'Generating {settings["num_samples"]} samples from proposal distribution.')
    # nde_sampler.run_sampler(num_samples=settings["num_samples"])

    # Step 2: Importance sample.
    #
    # Our target distribution is the posterior p(theta|d) = p(d|theta) * p(theta) / p(
    # d). For importance sampling, we need to evaluate the *unnormalized* posterior,
    # so we only need the likelihood p(d|theta) and the prior p(theta), but not the
    # evidence p(d).
    #
    # Sample from proposal distribution q(theta|d) and reweight the samples theta_i with
    #
    #       w_i = p(theta_i|d) / q(theta_i|d)
    #
    # to obtain weighted samples from the proposal distribution.

    # log_evidences = []
    # log_evidences_std = []
    # for idx in range(20):
    #     nde_sampler.run_sampler(num_samples=settings["num_samples"])
    #     nde_sampler.importance_sample(
    #         num_processes=settings.get("num_processes", 1),
    #         time_marginalization_kwargs=time_marginalization_kwargs,
    #         phase_marginalization_kwargs=phase_marginalization_kwargs,
    #     )
    #     log_evidences.append(nde_sampler.log_evidence)
    #     log_evidences_std.append(nde_sampler.log_evidence_std)
    # import numpy as np
    # log_evidences = np.array(log_evidences)
    # log_evidences_std = np.array(log_evidences_std)
    # print(np.std(log_evidences) / np.mean(log_evidences_std))

    if synthetic_phase:
        print(f"Sampling synthetic phase.")
        result.sample_synthetic_phase(synthetic_phase_kwargs)

    print(f"Importance sampling.")
    result.importance_sample(
        num_processes=settings.get("num_processes", 1),
        time_marginalization_kwargs=time_marginalization_kwargs,
        phase_marginalization_kwargs=phase_marginalization_kwargs,
        calibration_marginalization_kwargs=calibration_marginalization_kwargs,
    )
    result.print_summary()
    result.to_file(file_name=Path(args.outdir, "dingo_samples_weighted.hdf5"))

    # Diagnostics
    diagnostics_dir = join(args.outdir, "IS-diagnostics")
    if not exists(diagnostics_dir):
        makedirs(diagnostics_dir)
    print("Plotting diagnostics.")
    plot_diagnostics(
        result,
        diagnostics_dir,
        num_processes=settings.get("num_processes", 1),
        **settings.get("slice_plots", {}),
    )


if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd
import scipy
from sklearn.utils.extmath import randomized_svd
from dingo.core.dataset import DingoDataset


class SVDBasis(DingoDataset):

    dataset_type = "svd_basis"

    def __init__(
        self,
        file_name=None,
        dictionary=None,
    ):
        self.V = None
        self.Vh = None
        self.s = None
        self.n = None
        self.mismatches = None
        super().__init__(
            file_name=file_name,
            dictionary=dictionary,
            data_keys=["V", "s", "mismatches"],
        )

    def generate_basis(self, training_data: np.ndarray, n: int, method: str = "random"):
        """Generate the SVD basis from training data and store it.

        The SVD decomposition takes

        training_data = U @ diag(s) @ Vh

        where U and Vh are unitary.

        Parameters
        ----------
        training_data: np.ndarray
            Array of waveform data on the physical domain
        n: int
            Number of basis elements to keep.
            n=0 keeps all basis elements.
        method: str
            Select SVD method, 'random' or 'scipy'
        """
        if method == "random":
            if n == 0:
                n = min(training_data.shape)

            U, s, Vh = randomized_svd(training_data, n, random_state=0)

            self.Vh = Vh.astype(np.complex128)  # TODO: fix types
            self.V = self.Vh.T.conj()
            self.n = n
            self.s = s
        elif method == "scipy":
            # Code below uses scipy's svd tool. Likely slower.
            U, s, Vh = scipy.linalg.svd(training_data, full_matrices=False)
            V = Vh.T.conj()

            if (n == 0) or (n > len(V)):
                self.V = V
                self.Vh = Vh
            else:
                self.V = V[:, :n]
                self.Vh = Vh[:n, :]

            self.n = len(self.Vh)
            self.s = s
        else:
            raise ValueError(f"Unsupported SVD method: {method}.")

    def compute_test_mismatches(
        self,
        data: np.ndarray,
        parameters: pd.DataFrame = None,
        increment: int = 50,
        verbose: bool = False,
    ):
        """
        Test SVD basis by computing mismatches of compressed / decompressed data
        against original data. Results are saved as a DataFrame.

        Parameters
        ----------
        data : np.ndarray
            Array of data sets to validate against.
        parameters : pd.DataFrame
            Optional labels for the data sets. This is useful for checking performance on
            particular regions of the parameter space.
        increment : int
            Specifies SVD truncations for computing mismatches. E.g., increment = 50
            means that the SVD will be truncated at size [50, 100, 150, ..., len(data)].
        verbose : bool
            Whether to print summary statistics.
        """
        if len(data) != len(parameters):
            raise ValueError(
                f"Incompatible data: len(data) == {len(data)} and len("
                f"parameters) == {len(parameters)} do not match."
            )
        if parameters is not None:
            self.mismatches = parameters.copy()
        else:
            self.mismatches = pd.DataFrame()

        for n in np.append(np.arange(increment, self.n, increment), self.n):
            mismatches = np.empty(len(data))
            for i, d in enumerate(data):
                compressed = d @ self.V[:, :n]
                reconstructed = compressed @ self.Vh[:n]
                norm1 = np.sqrt(np.sum(np.abs(d) ** 2))
                norm2 = np.sqrt(np.sum(np.abs(reconstructed) ** 2))
                inner = np.sum(d.conj() * reconstructed).real
                mismatches[i] = 1 - inner / (norm1 * norm2)
            self.mismatches[f"mismatch n={n}"] = mismatches

        if verbose:
            self.print_validation_summary()

    def print_validation_summary(self):
        """
        Print a summary of the validation mismatches.
        """
        if self.mismatches is not None:
            for col in self.mismatches:
                if "mismatch" in col:
                    n = int(col.split(sep="=")[-1])
                    mismatches = self.mismatches[col]
                    print(f"n = {n}")
                    print("  Mean mismatch = {}".format(np.mean(mismatches)))
                    print("  Standard deviation = {}".format(np.std(mismatches)))
                    print("  Max mismatch = {}".format(np.max(mismatches)))
                    print("  Median mismatch = {}".format(np.median(mismatches)))
                    print("  Percentiles:")
                    print("    99    -> {}".format(np.percentile(mismatches, 99)))
                    print("    99.9  -> {}".format(np.percentile(mismatches, 99.9)))
                    print("    99.99 -> {}".format(np.percentile(mismatches, 99.99)))

    def decompress(self, coefficients: np.ndarray):
        """
        Convert from basis coefficients back to raw data representation.

        Parameters
        ----------
        coefficients : np.ndarray
            Array of basis coefficients

        Returns
        -------
        array of decompressed data
        """
        return coefficients @ self.Vh

    def compress(self, data: np.ndarray):
        """
        Convert from data (e.g., frequency series) to compressed representation in
        terms of basis coefficients.

        Parameters
        ----------
        data : np.ndarray

        Returns
        -------
        array of basis coefficients
        """
        return data @ self.V

    def from_file(self, filename):
        """
        Load the SVD basis from a HDF5 file.

        Parameters
        ----------
        filename : str
        """
        super().from_file(filename)
        if self.V is None:
            raise KeyError("File does not contain SVD V matrix. No SVD basis to load.")
        self.Vh = self.V.T.conj()
        self.n = self.V.shape[1]

    def from_dictionary(self, dictionary: dict):
        """
        Load the SVD basis from a dictionary.

        Parameters
        ----------
        dictionary : dict
            The dictionary should contain at least a 'V' key, and optionally an 's' key.
        """
        super().from_dictionary(dictionary)
        if self.V is None:
            raise KeyError("dict does not contain SVD V matrix. No SVD basis to load.")
        self.Vh = self.V.T.conj()
        self.n = self.V.shape[1]

    # def truncate(self, n: int):
    #     """
    #     Truncate size of SVD.
    #
    #     Parameters
    #     ----------
    #     n : int
    #         New SVD size. Should be less than current size.
    #     """
    #     if n > self.n or n < 0:
    #         print(f"Cannot truncate SVD from size n={self.n} to n={n}.")
    #     else:
    #         self.V = self.V[:, :n]
    #         self.Vh = self.Vh[:n, :]
    #         self.s = self.s[:n]
    #         self.n = n


class ApplySVD(object):
    """Transform operator for applying an SVD compression / decompression."""

    def __init__(self, svd_basis: SVDBasis, inverse: bool = False):
        """
        Parameters
        ----------
        svd_basis : SVDBasis
        inverse : bool
            Whether to apply for the forward (compression) or inverse (decompression)
            transform. Default: False.
        """
        self.svd_basis = svd_basis
        self.inverse = inverse

    def __call__(self, waveform: dict):
        """
        Parameters
        ----------
        waveform : dict
            Values should be arrays containing waveforms to be transformed.

        Returns
        -------
        dict of the same form as the input, but with transformed waveforms.
        """
        if not self.inverse:
            func = self.svd_basis.compress
        else:
            func = self.svd_basis.decompress
        return {k: func(v) for k, v in waveform.items()}



import torch
import argparse
from dingo.gw.result import Result

"""
2. March 2023: Previously, we stored the input_dim of the embedding net as a tuple, 
e.g., (2, 3, 8033). This is created in dingo.core.nn.nsf.autocomplete_model_kwargs_nsf, 
where we previously set:

    model_kwargs["embedding_net_kwargs"]["input_dims"] = data_sample[1].shape
    
This is problematic when saving the metadata as string in hdf5, since python does not 
properly convert the tuple to a string. This results in corrupted output with dingo_ls,

          input_dims: !!python/tuple
          - 2
          - 3
          - 8033
          
and causes trouble when interfacing with pesummary. In commit 
53929ae0ce5eee0607510119194167cf9db003a4, we thus changed the input_dim to be stored as 
a list:

    model_kwargs["embedding_net_kwargs"]["input_dims"] = list(data_sample[1].shape)
    
Old models and result files (pre commit 53929ae0ce5eee0607510119194167cf9db003a4) need 
to be updated with the present script to account for this change (note: training will 
still work, updates to old models are only required when running inference and 
interfacing with pesummary). 
"""


def parse_args():

    parser = argparse.ArgumentParser()
    parser.add_argument("--checkpoint", type=str, default=None)
    parser.add_argument("--results_file", type=str, default=None)
    return parser.parse_args()


def change_dict_elements_recursively(d, target_key, f, previous_keys=()):
    for k, el in d.items():
        if k == target_key:
            print(f"Changing {previous_keys + (k,)} from {el} to {f(el)}.")
            d[k] = f(el)
        elif type(el) == dict:
            change_dict_elements_recursively(el, target_key, f, previous_keys + (k,))


def main():
    args = parse_args()

    if args.checkpoint is not None:
        print(f"Updating model {args.checkpoint}.")
        # Typically training is done on the GPU, so the model could be saved on a GPU
        # device. Since this routine may be run on a CPU machine, allow for a remap of the
        # torch tensors.
        if torch.cuda.is_available():
            d = torch.load(args.checkpoint)
        else:
            d = torch.load(args.checkpoint, map_location=torch.device("cpu"))

        d["model_kwargs"]["embedding_net_kwargs"]["input_dims"] = list(
            d["model_kwargs"]["embedding_net_kwargs"]["input_dims"]
        )

        torch.save(d, args.checkpoint)

    if args.results_file is not None:
        print(f"Updating results file {args.results_file}.")
        res = Result(file_name=args.results_file)
        change_dict_elements_recursively(res.settings, "input_dims", list)
        res.to_file(file_name=args.results_file)


if __name__ == "__main__":
    main()



import copy
import textwrap
import yaml
import argparse
from multiprocessing import Pool

import numpy as np
import pandas as pd
from typing import Tuple, Dict
from threadpoolctl import threadpool_limits
from torchvision.transforms import Compose
from bilby.gw.prior import BBHPriorDict

from dingo.gw.dataset.waveform_dataset import WaveformDataset
from dingo.gw.prior import build_prior_with_defaults
from dingo.gw.domains import build_domain
from dingo.gw.transforms import WhitenFixedASD
from dingo.gw.waveform_generator import (
    WaveformGenerator,
    NewInterfaceWaveformGenerator,
    generate_waveforms_parallel,
)
from dingo.gw.SVD import SVDBasis, ApplySVD


def generate_parameters_and_polarizations(
    waveform_generator: WaveformGenerator,
    prior: BBHPriorDict,
    num_samples: int,
    num_processes: int,
) -> Tuple[pd.DataFrame, Dict[str, np.ndarray]]:
    """
    Generate a dataset of waveforms based on parameters drawn from the prior.

    Parameters
    ----------
    waveform_generator : WaveformGenerator
    prior : Prior
    num_samples : int
    num_processes : int

    Returns
    -------
    pandas DataFrame of parameters
    dictionary of numpy arrays corresponding to waveform polarizations
    """
    print("Generating dataset of size " + str(num_samples))
    parameters = pd.DataFrame(prior.sample(num_samples))

    if num_processes > 1:
        with threadpool_limits(limits=1, user_api="blas"):
            with Pool(processes=num_processes) as pool:
                polarizations = generate_waveforms_parallel(
                    waveform_generator, parameters, pool
                )
    else:
        polarizations = generate_waveforms_parallel(waveform_generator, parameters)

    # Find cases where waveform generation failed and only return data for successful ones
    wf_failed = np.any(np.isnan(polarizations["h_plus"]), axis=1)
    if wf_failed.any():
        idx_failed = np.where(wf_failed)[0]
        idx_ok = np.where(~wf_failed)[0]
        polarizations_ok = {k: v[idx_ok] for k, v in polarizations.items()}
        parameters_ok = parameters.iloc[idx_ok]
        failed_percent = 100 * len(idx_failed) / len(parameters)
        print(
            f"{len(idx_failed)} out of {len(parameters)} configuration ({failed_percent:.1f}%) failed to generate."
        )
        with pd.option_context("display.max_rows", None, "display.max_columns", None):
            print(parameters.iloc[idx_failed])
        print(
            f"Only returning the {len(idx_ok)} successfully generated configurations."
        )
        return parameters_ok, polarizations_ok

    return parameters, polarizations


def train_svd_basis(dataset: WaveformDataset, size: int, n_train: int):
    """
    Train (and optionally validate) an SVD basis.

    Parameters
    ----------
    dataset : WaveformDataset
        Contains waveforms to be used for building SVD.
    size : int
        Number of elements to keep for the SVD basis.
    n_train : int
        Number of training waveforms to use. Remaining are used for validation. Note
        that the actual number of training waveforms is n_train * len(polarizations),
        since there is one waveform used for each polarization.

    Returns
    -------
    SVDBasis, n_train, n_test
        Since EOB waveforms can fail to generate, provide also the number used in
        training and validation.
    """
    # Prepare data for training and validation.
    train_data = np.vstack([val[:n_train] for val in dataset.polarizations.values()])
    test_data = np.vstack([val[n_train:] for val in dataset.polarizations.values()])
    test_parameters = pd.concat(
        [
            # I would like to save the polarization, but saving the dataframe with
            # string columns causes problems. Fix this later.
            # dataset.parameters.iloc[n_train:].assign(polarization=pol)
            dataset.parameters.iloc[n_train:]
            for pol in dataset.polarizations
        ]
    )
    test_parameters.reset_index(drop=True, inplace=True)

    print("Building SVD basis.")
    basis = SVDBasis()
    basis.generate_basis(train_data, size)

    assert np.allclose(basis.V[: dataset.domain.min_idx], 0)

    # Since there is a possibility that the size of the dataset returned by
    # generate_parameters_and_polarizations is smaller than requested, we don't assume
    # that there are n_test samples. Instead we just look at the size of the test
    # dataset.
    if test_data.size != 0:
        basis.compute_test_mismatches(
            test_data, parameters=test_parameters, verbose=True
        )

    # Return also the true number of samples. Some EOB waveforms may have failed to
    # generate, so this could be smaller than the number requested.
    n_ifos = len(dataset.polarizations)
    n_train = len(train_data) // n_ifos
    n_test = len(test_data) // n_ifos

    return basis, n_train, n_test


def generate_dataset(settings: Dict, num_processes: int) -> WaveformDataset:
    """
    Generate a waveform dataset.

    Parameters
    ----------
    settings : dict
        Dictionary of settings to configure the dataset
    num_processes : int

    Returns
    -------
    A WaveformDataset based on the settings.
    """

    prior = build_prior_with_defaults(settings["intrinsic_prior"])
    domain = build_domain(settings["domain"])

    new_interface_flag = settings["waveform_generator"].get("new_interface", False)
    if new_interface_flag:
        waveform_generator = NewInterfaceWaveformGenerator(
            domain=domain,
            **settings["waveform_generator"],
        )
    else:
        waveform_generator = WaveformGenerator(
            domain=domain,
            **settings["waveform_generator"],
        )

    dataset_dict = {"settings": settings}

    if "compression" in settings:
        compression_transforms = []

        if "whitening" in settings["compression"]:
            compression_transforms.append(
                WhitenFixedASD(
                    domain,
                    asd_file=settings["compression"]["whitening"],
                    inverse=False,
                )
            )

        if "svd" in settings["compression"]:
            svd_settings = settings["compression"]["svd"]

            # Load an SVD basis from file, if specified.
            if "file" in svd_settings:
                basis = SVDBasis(file_name=svd_settings["file"])

            # Otherwise, generate the basis based on simulated waveforms.
            else:
                # If using whitened waveforms, then the SVD should be based on these.
                waveform_generator.transform = Compose(compression_transforms)

                n_train = svd_settings["num_training_samples"]
                n_test = svd_settings.get("num_validation_samples", 0)
                parameters, polarizations = generate_parameters_and_polarizations(
                    waveform_generator,
                    prior,
                    n_train + n_test,
                    num_processes,
                )
                svd_dataset_settings = copy.deepcopy(settings)
                svd_dataset_settings["num_samples"] = len(parameters)
                del svd_dataset_settings["compression"]["svd"]

                # We build a WaveformDataset containing the SVD-training waveforms
                # because when constructed, it will automatically zero the waveforms
                # below f_min. This is useful for EOB waveforms, which are Fourier
                # transformed from time domain, and hence are nonzero below f_min. The
                # waveforms need to be zeroed below f_min because this corresponds to
                # setting the lower bound of the likelihood integral.

                svd_dataset = WaveformDataset(
                    dictionary={
                        "parameters": parameters,
                        "polarizations": polarizations,
                        "settings": svd_dataset_settings,
                    }
                )
                basis, n_train, n_test = train_svd_basis(
                    svd_dataset, svd_settings["size"], n_train
                )
                # Reset the true number of samples, in case this has changed due to
                # failure to generate some EOB waveforms.
                svd_settings["num_training_samples"] = n_train
                svd_settings["num_validation_samples"] = n_test

            compression_transforms.append(ApplySVD(basis))
            dataset_dict["svd"] = basis.to_dictionary()

        waveform_generator.transform = Compose(compression_transforms)

    # Generate main dataset
    parameters, polarizations = generate_parameters_and_polarizations(
        waveform_generator, prior, settings["num_samples"], num_processes
    )
    dataset_dict["parameters"] = parameters
    dataset_dict["polarizations"] = polarizations
    # Update to take into account potentially failed configurations
    dataset_dict[settings["num_samples"]] = len(parameters)

    dataset = WaveformDataset(dictionary=dataset_dict)
    return dataset


def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent(
            """\
        Generate a waveform dataset based on a settings file.
        """
        ),
    )
    parser.add_argument(
        "--settings_file",
        type=str,
        required=True,
        help="YAML file containing database settings",
    )
    parser.add_argument(
        "--num_processes",
        type=int,
        default=1,
        help="Number of processes to use in pool for parallel waveform generation",
    )
    parser.add_argument(
        "--out_file",
        type=str,
        default="waveform_dataset.hdf5",
        help="Name of file for storing dataset.",
    )
    return parser.parse_args()


def main():
    args = parse_args()

    # Load settings
    with open(args.settings_file, "r") as f:
        settings = yaml.safe_load(f)

    dataset = generate_dataset(settings, args.num_processes)
    dataset.to_file(args.out_file)


if __name__ == "__main__":
    main()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from importlib.metadata import version as _version
sys.path.insert(0, os.path.abspath('../../dingo'))


# -- Project information -----------------------------------------------------

project = 'dingo-gw'
copyright = '2022-2023, Stephen Green'
author = 'Stephen Green'

# The full version, including alpha/beta/rc tags
version = _version('dingo-gw')


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'myst_nb',
    'sphinxcontrib.mermaid',
    'sphinxcontrib.bibtex',
    'sphinx_math_dollar',
]

myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "strikethrough",
    "substitution",
    "tasklist",
]

mathjax_config = {
    'tex2jax': {
        'inlineMath': [ ["\\(","\\)"] ],
        'displayMath': [["\\[","\\]"] ],
    },
}

mathjax3_config = {
  "tex": {
    "inlineMath": [['\\(', '\\)']],
    "displayMath": [["\\[", "\\]"]],
  }
}

myst_heading_anchors = 2

bibtex_bibfiles = ['refs.bib']
bibtex_default_style = 'unsrt'

# Include the __init__ docstring with the autoclass
autoclass_content = 'both'

# Napolean Settings
napoleon_google_docstring = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']



import pytest
import numpy as np

from dingo.gw.domains import build_domain
from dingo.gw.gwutils import get_mismatch
from dingo.gw.injection import GWSignal
from dingo.gw.waveform_generator import sum_contributions_m


@pytest.fixture
def signal_setup_EOB():
    wfg_kwargs = {
        "approximant": "SEOBNRv4PHM",
        "f_ref": 10.0,
        "f_start": 10.0,
        "spin_conversion_phase": 0,
    }
    wfg_domain = build_domain(
        {
            "type": "FrequencyDomain",
            "f_min": 20,
            "f_max": 1024,
            "delta_f": 0.125,
            "window_factor": 1.0,
        }
    )
    data_domain = build_domain(
        {
            "type": "FrequencyDomain",
            "f_min": 20.0,
            "f_max": 1024.0,
            "delta_f": 0.125,
            "window_factor": 0.9374713897717841,
        }
    )
    ifo_list = ["H1", "L1"]
    t_ref = 1126259462.4
    signal = GWSignal(wfg_kwargs, wfg_domain, data_domain, ifo_list, t_ref)
    return signal


@pytest.fixture
def BBH_parameters():
    parameters = {
        "mass_1": 60.29442201204798,
        "mass_2": 25.460299253933126,
        "phase": 2.346269257440926,
        "a_1": 0.07104636316747037,
        "a_2": 0.7853578509086726,
        "tilt_1": 1.8173336549500292,
        "tilt_2": 0.4380213394743055,
        "phi_12": 5.892609139936818,
        "phi_jl": 1.6975651971466297,
        "theta_jn": 1.0724395559873239,
        "luminosity_distance": 100.0,
        "geocent_time": 0.0,
        "ra": 1.0,
        "dec": 2.0,
        "psi": 2.5,
    }
    return parameters


def test_signal_m_EOB(signal_setup_EOB, BBH_parameters):
    signal = signal_setup_EOB
    p = BBH_parameters
    domain = signal.waveform_generator.domain

    phase_shift = np.random.uniform(high=2 * np.pi)
    waveform_ref = signal.signal({**p, "phase": p["phase"] + phase_shift})["waveform"]
    waveform_m = signal.signal_m(p)
    waveform_m = {k: v["waveform"] for k, v in waveform_m.items()}
    waveform = sum_contributions_m(waveform_m, phase_shift=phase_shift)
    mismatches = [
        get_mismatch(
            waveform_ref[k],
            waveform[k],
            domain,
            asd_file="aLIGO_ZERO_DET_high_P_asd.txt",
        )
        for k in waveform.keys()
    ]
    assert np.max(mismatches) < 1e-4



import copy
from typing import Dict, Union
import numpy as np
import torch.utils.data
from torchvision.transforms import Compose

from dingo.core.dataset import DingoDataset
from dingo.gw.SVD import SVDBasis, ApplySVD
from dingo.gw.domains import build_domain
from dingo.gw.transforms import WhitenFixedASD


class WaveformDataset(DingoDataset, torch.utils.data.Dataset):
    """This class stores a dataset of waveforms (polarizations) and corresponding
    parameters.

    It can load the dataset either from an HDF5 file or suitable dictionary.

    Once a waveform data set is in memory, the waveform data are consumed through a
    __getitem__() call, optionally applying a chain of transformations, which are classes
    that implement a __call__() method.
    """

    dataset_type = "waveform_dataset"

    def __init__(
        self,
        file_name=None,
        dictionary=None,
        transform=None,
        precision=None,
        domain_update=None,
        svd_size_update=None,
    ):
        """
        For constructing, provide either file_name, or dictionary containing data and
        settings entries, or neither.

        Parameters
        ----------
        file_name : str
            HDF5 file containing a dataset
        dictionary : dict
            Contains settings and data entries. The dictionary keys should be
            'settings', 'parameters', and 'polarizations'.
        transform : Transform
            Transform to be applied to dataset samples when accessed through __getitem__
        precision : str ('single', 'double')
            If provided, changes precision of loaded dataset.
        domain_update : dict
            If provided, update domain from existing domain using new settings.
        svd_size_update : int
            If provided, reduces the SVD size when decompressing (for speed).
        """
        self.domain = None
        self.transform = transform
        self.decompression_transform = None
        self.precision = precision
        super().__init__(
            file_name=file_name,
            dictionary=dictionary,
            data_keys=["parameters", "polarizations", "svd"],
        )

        if (
            self.parameters is not None
            and self.polarizations is not None
            and self.settings is not None
        ):
            self.load_supplemental(domain_update, svd_size_update)

    def load_supplemental(self, domain_update=None, svd_size_update=None):
        """Method called immediately after loading a dataset.

        Creates (and possibly updates) domain, updates dtypes, and initializes any
        decompression transform. Also zeros data below f_min, and truncates above f_max.

        Parameters
        ----------
        domain_update : dict
            If provided, update domain from existing domain using new settings.
        svd_size_update : int
            If provided, reduces the SVD size when decompressing (for speed).
        """
        self.domain = build_domain(self.settings["domain"])

        # We always call update_domain() (even if domain_update is None) because we
        # want to be sure that the data are consistent with the saved settings. In
        # particular, this zeroes the waveforms for f < f_min.
        self.update_domain(domain_update)

        # Update dtypes if necessary
        if self.precision is not None:
            if self.precision == "single":
                complex_type = np.complex64
                real_type = np.float32
            elif self.precision == "double":
                complex_type = np.complex128
                real_type = np.float64
            else:
                raise TypeError(
                    'precision can only be changed to "single" or "double".'
                )
            self.parameters = self.parameters.astype(real_type, copy=False)
            for k, v in self.polarizations.items():
                self.polarizations[k] = v.astype(complex_type, copy=False)

            # This should probably be moved to the SVDBasis class.
            if self.svd is not None:
                self.svd["V"] = self.svd["V"].astype(complex_type, copy=False)
                # For backward compatibility; in future, this will be there.
                if "s" in self.svd:
                    self.svd["s"] = self.svd["s"].astype(real_type, copy=False)

        if self.settings.get("compression", None) is not None:
            self.initialize_decompression(svd_size_update)

    def update_domain(self, domain_update: dict = None):
        """
        Update the domain based on new configuration.

        The waveform dataset provides waveform polarizations in a particular domain. In
        Frequency domain, this is [0, domain._f_max]. Furthermore, data is set to 0 below
        domain._f_min. In practice one may want to train a network based on  slightly
        different domain settings, which corresponds to truncating the likelihood
        integral.

        This method provides functionality for that. It truncates and/or zeroes the
        dataset to the range specified by the domain, by calling domain.update_data.

        Parameters
        ----------
        domain_update : dict
            Settings dictionary. Must contain a subset of the keys contained in
            domain_dict.
        """
        if domain_update is not None:
            self.domain.update(domain_update)

        # Determine where any domain adjustment must be applied. If the dataset is SVD
        # compressed, then adjust the SVD matrices. Otherwise, adjust the dataset
        # itself.
        if (
            self.settings.get("compression", None) is not None
            and "svd" in self.settings["compression"]
        ):
            self.svd["V"] = self.domain.update_data(self.svd["V"], axis=0)
        else:
            for k, v in self.polarizations.items():
                self.polarizations[k] = self.domain.update_data(v)

    def initialize_decompression(self, svd_size_update: int = None):
        """
        Sets up decompression transforms. These are applied to the raw dataset before
        self.transform. E.g., SVD decompression.

        Parameters
        ----------
        svd_size_update : int
            If provided, reduces the SVD size when decompressing (for speed).
        """
        decompression_transform_list = []

        # These transforms must be in reverse order compared to when dataset was
        # constructed.

        if "svd" in self.settings["compression"]:
            assert self.svd is not None

            # We allow the option to reduce the size of the SVD used for decompression,
            # since decompression is the costliest preprocessing operation. Be careful
            # when using this to not introduce a large mismatch.
            if svd_size_update is not None:
                if svd_size_update > self.svd["V"].shape[-1] or svd_size_update < 0:
                    raise ValueError(
                        f"Cannot truncate SVD from size "
                        f"{self.svd['V'].shape[-1]} to size "
                        f"{svd_size_update}."
                    )
                self.svd["V"] = self.svd["V"][:, :svd_size_update]
                self.svd["s"] = self.svd["s"][:svd_size_update]
                for k, v in self.polarizations.items():
                    self.polarizations[k] = v[:, :svd_size_update]

            svd_basis = SVDBasis(dictionary=self.svd)
            decompression_transform_list.append(ApplySVD(svd_basis, inverse=True))

        if "whitening" in self.settings["compression"]:
            decompression_transform_list.append(
                WhitenFixedASD(
                    self.domain,
                    asd_file=self.settings["compression"]["whitening"],
                    inverse=True,
                    precision=self.precision,
                )
            )

        self.decompression_transform = Compose(decompression_transform_list)

    def __len__(self):
        """The number of waveform samples."""
        return len(self.parameters)

    def __getitem__(self, idx) -> Dict[str, Dict[str, Union[float, np.ndarray]]]:
        """
        Return a nested dictionary containing parameters and waveform polarizations
        for sample with index `idx`. If defined, a chain of transformations is applied to
        the waveform data.
        """
        parameters = self.parameters.iloc[idx].to_dict()
        polarizations = {
            pol: waveforms[idx] for pol, waveforms in self.polarizations.items()
        }

        # Decompression transforms are assumed to apply only to the waveform,
        # and do not involve parameters.
        if self.decompression_transform is not None:
            polarizations = self.decompression_transform(polarizations)

        # Main transforms can depend also on parameters.
        data = {"parameters": parameters, "waveform": polarizations}
        if self.transform is not None:
            data = self.transform(data)
        return data

    def parameter_mean_std(self):
        mean = self.parameters.mean().to_dict()
        std = self.parameters.std().to_dict()
        return mean, std



#
#  Modified version of bilby_pipe parser.
#

import argparse

import configargparse
from bilby_pipe.bilbyargparser import BilbyArgParser
from bilby_pipe.utils import (
    get_version_information,
    logger,
    nonefloat,
    noneint,
    nonestr,
)

logger.name = "dingo_pipe"


class StoreBoolean(argparse.Action):
    """argparse class for robust handling of booleans with configargparse

    When using configargparse, if the argument is setup with
    action="store_true", but the default is set to True, then there is no way,
    in the config file to switch the parameter off. To resolve this, this class
    handles the boolean properly.

    """

    def __call__(self, parser, namespace, value, option_string=None):
        value = str(value).lower()
        if value in ["true"]:
            setattr(namespace, self.dest, True)
        else:
            setattr(namespace, self.dest, False)


def create_parser(top_level=True):
    """Creates the BilbyArgParser for dingo_pipe

    Parameters
    ----------
    top_level:
        If true, parser is to be used at the top-level with requirement
        checking etc., else it is an internal call and will be ignored.

    Returns
    -------
    parser: BilbyArgParser instance
        Argument parser

    """
    parser = BilbyArgParser(
        usage="Perform inference with dingo based on a .ini file.",
        ignore_unknown_config_file_keys=False,
        allow_abbrev=False,
        formatter_class=configargparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add("ini", type=str, is_config_file=True, help="Configuration ini file")
    parser.add("-v", "--verbose", action="store_true", help="Verbose output")
    # parser.add(
    #     "--version",
    #     action="version",
    #     version=f"%(prog)s={get_version_information()}\nbilby={bilby.__version__}",
    # )

    calibration_parser = parser.add_argument_group(
        "Calibration arguments",
        description="Which calibration model and settings to use. Calibration "
                    "uncertainty is marginalizaed over during importance sampling.",
    )
    calibration_parser.add(
        "--calibration-model",
        type=nonestr,
        default=None,
        choices=["CubicSpline", None],
        help="Choice of calibration model, if None, no calibration is used",
    )

    calibration_parser.add(
        "--spline-calibration-envelope-dict",
        type=nonestr,
        default=None,
        help=("Dictionary pointing to the spline calibration envelope files"),
    )

    calibration_parser.add(
        "--spline-calibration-nodes",
        type=int,
        default=10,
        help=("Number of calibration nodes"),
    )

    calibration_parser.add(
        "--spline-calibration-curves",
        type=int,
        default=1000,
        help=("Number of calibration curves to use in marginalizing over calibration "
              "uncertainty"),
    )
    #
    # calibration_parser.add(
    #     "--spline-calibration-amplitude-uncertainty-dict",
    #     type=nonestr,
    #     default=None,
    #     help=(
    #         "Dictionary of the amplitude uncertainties for the constant "
    #         "uncertainty model"
    #     ),
    # )
    #
    # calibration_parser.add(
    #     "--spline-calibration-phase-uncertainty-dict",
    #     type=nonestr,
    #     default=None,
    #     help=(
    #         "Dictionary of the phase uncertainties for the constant "
    #         "uncertainty model"
    #     ),
    # )
    # calibration_parser.add(
    #     "--calibration-prior-boundary",
    #     type=nonestr,
    #     default="reflective",
    #     help="Boundary methods for the calibration prior boundary",
    # )

    if top_level is False:
        parser.add("--idx", type=int, help="The level A job index", default=0)
        parser.add(
            "--data-dump-file",
            type=nonestr,
            default=None,
            help="Filename for the data dump: only used internally by data_analysis",
        )
        parser.add(
            "--event-data-file",
            type=nonestr,
            default=None,
            help="Filename for the event: only used internally by sampling and "
            "importance_sampling",
        )
        parser.add(
            "--proposal-samples-file",
            type=nonestr,
            default=None,
            help="Filename for the proposal samples: only used internally by "
            "importance_sampling",
        )
        parser.add(
            "--importance-sampling-generation",
            action="store_true",
            help="Whether to prepare data based on the updated importance sampling "
                 "settings rather than network settings. This is used internally for "
                 "data generation, when preparing different data for the importance "
                 "sampling stage.",
        )

    data_gen_pars = parser.add_argument_group(
        "Data generation arguments",
        description="How to generate the data, e.g., from a list of gps times or simulated Gaussian noise.",
    )

    data_gen_pars.add(
        "--ignore-gwpy-data-quality-check",
        action=StoreBoolean,
        default=True,
        help=(
            "Ignores the check to see if data queried from GWpy (ie not gaussian "
            "noise) is obtained from time when the IFOs are in science mode."
        ),
    )
    #
    # data_gen_pars.add(
    #     "--gps-tuple",
    #     type=nonestr,
    #     help=(
    #         "Tuple of the (start, step, number) of GPS start times. For"
    #         " example, (10, 1, 3) produces the gps start times [10, 11, 12]."
    #         " If given, gps-file is ignored."
    #     ),
    #     default=None,
    # )
    # data_gen_pars.add(
    #     "--gps-file",
    #     type=nonestr,
    #     help=(
    #         "File containing segment GPS start times. This can be a multi-"
    #         "column file if (a) it is comma-separated and (b) the zeroth "
    #         "column contains the gps-times to use"
    #     ),
    #     default=None,
    # )
    data_gen_pars.add(
        "--timeslide-file",
        type=nonestr,
        help=(
            "File containing detector timeslides. "
            "Requires a GPS time file to also be provided. One column for each "
            "detector. Order of detectors specified by `--detectors` argument. "
            "Number of timeslides must correspond to the number of GPS times provided."
        ),
        default=None,
    )
    # data_gen_pars.add(
    #     "--timeslide-dict",
    #     type=nonestr,
    #     help=(
    #         "Dictionary containing detector timeslides: applies a fixed offset"
    #         " per detector. E.g. to apply +1s in H1, {H1: 1}"
    #     ),
    #     default=None,
    # )
    data_gen_pars.add(
        "--trigger-time",
        default=None,
        type=nonestr,
        help=(
            "Either a GPS trigger time, or the event name (e.g. GW150914). "
            "For event names, the gwosc package is used to identify the "
            "trigger time"
        ),
    )
    # data_gen_pars.add(
    #     "--n-simulation",
    #     type=int,
    #     default=0,
    #     help=(
    #         "Number of simulated segments to use with gaussian-noise "
    #         "Note, this must match the number of injections specified"
    #     ),
    # )
    data_gen_pars.add(
        "--data-dict",
        default=None,
        type=nonestr,
        help="Dictionary of paths to gwf, or hdf5 data files",
    )
    data_gen_pars.add(
        "--data-format",
        default=None,
        type=nonestr,
        help=(
            "If given, the data format to pass to "
            " `gwpy.timeseries.TimeSeries.read(), see "
            " gwpy.github.io/docs/stable/timeseries/io.html"
        ),
    )
    data_gen_pars.add(
        "--allow-tape",
        default=True,
        action=StoreBoolean,
        help=(
            "If true (default), allow reading data from tape."
            " See `gwpy.timeseries.TimeSeries.get() for more information."
        ),
    )
    data_gen_pars.add(
        "--channel-dict",
        type=nonestr,
        default=None,
        help=(
            "Channel dictionary: keys relate to the detector with values "
            "the channel name, e.g. 'GDS-CALIB_STRAIN'. For GWOSC open data, "
            "set the channel-dict keys to 'GWOSC'. Note, the "
            "dictionary should follow basic python dict syntax."
        ),
    )
    # data_type_pars = data_gen_pars.add_mutually_exclusive_group()
    # data_type_pars.add(
    #     "--gaussian-noise",
    #     action="store_true",
    #     help="If true, use simulated Gaussian noise",
    # )
    # data_type_pars.add(
    #     "--zero-noise",
    #     action="store_true",
    #     help="Use a zero noise realisation",
    # )

    det_parser = parser.add_argument_group(
        title="Detector arguments",
        description="How to set up the interferometers and power spectral density.",
    )
    # det_parser.add(
    #     "--coherence-test",
    #     action="store_true",
    #     help=(
    #         "Run the analysis for all detectors together and for each "
    #         "detector separately"
    #     ),
    # )
    det_parser.add(
        "--detectors",
        action="append",
        help=(
            "The names of detectors to use. If given in the ini file, "
            "detectors are specified by `detectors=[H1, L1]`. If given "
            "at the command line, as `--detectors H1 --detectors L1`"
        ),
    )
    det_parser.add(
        "--duration",
        type=nonefloat,
        # default=4,
        default=None,
        help="The duration of data around the event to use",
    )
    # det_parser.add(
    #     "--generation-seed",
    #     default=None,
    #     type=noneint,
    #     help=(
    #         "Random seed used during data generation. If no generation seed "
    #         "provided, a random seed between 1 and 1e6 is selected. If a seed "
    #         "is provided, it is used as the base seed and all generation jobs "
    #         "will have their seeds set as {generation_seed = base_seed + job_idx}."
    #     ),
    # )
    det_parser.add(
        "--psd-dict", type=nonestr, default=None, help="Dictionary of PSD files to use"
    )
    det_parser.add(
        "--psd-fractional-overlap",
        # default=0.5,
        default=0.0,
        type=float,
        help="Fractional overlap of segments used in estimating the PSD",
    )
    det_parser.add(
        "--post-trigger-duration",
        type=float,
        default=2.0,
        help=("Time (in s) after the trigger_time to the end of the segment"),
    )
    det_parser.add(
        "--sampling-frequency",
        # default=4096,
        default=None,
        type=nonefloat,
    )

    det_parser.add(
        "--psd-length",
        default=32,
        type=int,
        help=(
            "Sets the psd duration (up to the psd-duration-maximum). PSD "
            "duration calculated by psd-length x duration [s]. Default is 32."
        ),
    )
    det_parser.add(
        "--psd-maximum-duration",
        default=1024,
        type=int,
        help=("The maximum allowed PSD duration in seconds, default is 1024s."),
    )
    det_parser.add(
        "--psd-method",
        default="median",
        type=str,
        help="PSD method see gwpy.timeseries.TimeSeries.psd for options",
    )
    det_parser.add(
        "--psd-start-time",
        default=None,
        type=nonefloat,
        help=(
            "Start time of data (relative to the segment start) used to "
            " generate the PSD. Defaults to psd-duration before the"
            " segment start time"
        ),
    )
    det_parser.add(
        "--maximum-frequency",
        default=None,
        type=nonestr,
        help=(
            "The maximum frequency, given either as a float for all detectors "
            "or as a dictionary (see minimum-frequency)"
        ),
    )
    det_parser.add(
        "--minimum-frequency",
        # default="20",
        default=None,
        type=nonestr,
        help=(
            "The minimum frequency, given either as a float for all detectors "
            "or as a dictionary where all keys relate the detector with values"
            " of the minimum frequency, e.g. {H1: 10, L1: 20}. If the waveform"
            " generation should start the minimum frequency for any of the "
            "detectors, add another entry to the dictionary, e.g., "
            "{H1: 40, L1: 60, waveform: 20}."
        ),
    )
    det_parser.add(
        "--tukey-roll-off",
        # default=0.4,
        default=None,
        type=nonefloat,
        help="Roll off duration of tukey window in seconds, default is 0.4s",
    )
    det_parser.add(
        "--resampling-method",
        default="lal",
        type=str,
        choices=["lal", "gwpy"],
        help="Resampling method to use: lal matches the resampling used by lalinference/BayesWave",
    )

    injection_parser = parser.add_argument_group(
        title="Injection arguments",
        description="Whether to include software injections and how to generate them.",
    )
    # injection_parser.add(
    #     "--injection",
    #     action="store_true",
    #     default=False,
    #     help="Create data from an injection file",
    # )
    injection_parser_input = injection_parser.add_mutually_exclusive_group()
    injection_parser_input.add(
        "--injection-dict",
        type=nonestr,
        default=None,
        help="A single injection dictionary given in the ini file",
    )
    injection_parser_input.add(
        "--injection-file",
        type=nonestr,
        default=None,
        help=(
            "Injection file to use. See `bilby_pipe_create_injection_file --help`"
            " for supported formats"
        ),
    )
    # injection_parser.add(
    #     "--injection-numbers",
    #     action="append",
    #     type=nonestr,
    #     default=None,
    #     help=(
    #         "Specific injections rows to use from the injection_file, e.g. "
    #         "`injection_numbers=[0,3] selects the zeroth and third row. Can be "
    #         "a list of slice-syntax values, e.g, [0, 2:4] will produce [0, 2, 3]. "
    #         "Repeated entries will be ignored."
    #     ),
    # )
    # injection_parser.add(
    #     "--injection-waveform-approximant",
    #     type=nonestr,
    #     default=None,
    #     help="The name of the waveform approximant to use to create injections. "
    #     "If none is specified, then the `waveform-approximant` will be used"
    #     "as the `injection-waveform-approximant`.",
    # )
    # injection_parser.add(
    #     "--injection-waveform-arguments",
    #     type=nonestr,
    #     default=None,
    #     help=(
    #         "A dictionary of arbitrary additional waveform-arguments to pass "
    #         "to the bilby waveform generator's waveform arguments for the "
    #         "injection only"
    #     ),
    # )

    submission_parser = parser.add_argument_group(
        title="Job submission arguments",
        description="How the jobs should be formatted, e.g., which job scheduler to use.",
    )
    submission_parser.add(
        "--accounting",
        type=nonestr,
        help="Accounting group to use (see, https://accounting.ligo.org/user)",
    )
    submission_parser.add(
        "--accounting-user",
        type=nonestr,
        help="Accounting group user to use (see, https://accounting.ligo.org/user)",
    )
    submission_parser.add("--label", type=str, default="label", help="Output label")
    submission_parser.add(
        "--local",
        action="store_true",
        help="Run the job locally, i.e., not through a batch submission",
    )
    submission_parser.add(
        "--local-generation",
        action="store_true",
        help=(
            "Run the data generation job locally. This may be useful for "
            "running on a cluster where the compute nodes do not have "
            "internet access. For HTCondor, this is done using the local "
            "universe, for slurm, the jobs will be run at run-time"
        ),
    )
    submission_parser.add(
        "--local-plot", action="store_true", help="Run the plot job locally"
    )

    submission_parser.add(
        "--outdir",
        type=str,
        default="outdir",
        help="The output directory. If outdir already exists, an auto-incrementing naming scheme is used",
    )
    submission_parser.add(
        "--overwrite-outdir",
        action="store_true",
        help="If given, overwrite the outdir (if it exists)",
    )
    # submission_parser.add(
    #     "--periodic-restart-time",
    #     default=28800,
    #     type=int,
    #     help=(
    #         "Time after which the job will self-evict when scheduler=condor."
    #         " After this, condor will restart the job. Default is 28800."
    #         " This is used to decrease the chance of HTCondor hard evictions"
    #     ),
    # )
    submission_parser.add(
        "--request-disk",
        type=float,
        default=5,
        help="Disk allocation request in GB. Default is 5GB.",
    )
    submission_parser.add(
        "--request-memory",
        type=float,
        default=8.0,
        help="Memory allocation request (GB). Default is 8GB",
    )
    submission_parser.add(
        "--request-memory-generation",
        type=nonefloat,
        # default=None,
        default=8.0,  # Change this to None if possibility of ROQ (likely remove ROQ)
        help="Memory allocation request (GB) for data generation step",
    )
    submission_parser.add(
        "--request-cpus",
        type=int,
        default=1,
        help=(
            "Use multi-processing. This options sets the number of cores to "
            "request. To use a pool of 8 threads on an 8-core CPU, set "
            "request-cpus=8. For the dynesty, ptemcee, cpnest, and "
            "bilby_mcmc samplers, no additional sampler-kwargs are required"
        ),
    )
    submission_parser.add(
        "--request-cpus-importance-sampling",
        type=int,
        default=1,
        help=(
            "Use multi-processing. This options sets the number of cores to "
            "request per job when performing importance sampling. To use a pool of 8 "
            "threads on an 8-core CPU, set request-cpus-importance-sampling=8."
        ),
    )
    submission_parser.add(
        "--sampling-requirements",
        action="append",
        help=(
            "List of extra requirements for submitting sampling. Can be used to specify "
            "GPU memory, e.g., [TARGET.CUDAGlobalMemoryMb>40000]."
        ),
    )
    submission_parser.add(
        "--extra-lines",
        action="append",
        help=(
            "List of additional lines to include for all HTCondor submissions."
        ),
    )
    submission_parser.add(
        "--simple-submission",
        action="store_true",
        help=(
            "Strip off the following lines from submission files: getenv, universe, "
            "accounting_group, priority."
        ),
    )
    submission_parser.add(
        "--conda-env",
        type=nonestr,
        default=None,
        help="Either a conda environment name of a absolute path to the conda env folder.",
    )
    submission_parser.add(
        "--scheduler",
        type=str,
        default="condor",
        help="Format submission script for specified scheduler. Currently implemented: SLURM",
    )
    submission_parser.add(
        "--scheduler-args",
        type=nonestr,
        default=None,
        help=(
            "Space-separated #SBATCH command line args to pass to slurm. "
            "The args needed will depend on the setup of your slurm scheduler."
            "Please consult documentation for your local cluster (slurm only)."
        ),
    )
    submission_parser.add(
        "--scheduler-module",
        type=nonestr,
        action="append",
        default=None,
        help="Space-separated list of modules to load at runtime (slurm only)",
    )
    submission_parser.add(
        "--scheduler-env",
        type=nonestr,
        default=None,
        help="Python environment to activate (slurm only)",
    )
    submission_parser.add(
        "--scheduler-analysis-time", type=nonestr, default="7-00:00:00", help=""
    )
    submission_parser.add(
        "--submit",
        action="store_true",
        help="Attempt to submit the job after the build",
    )
    submission_parser.add(
        "--condor-job-priority",
        type=int,
        default=0,
        help=(
            "Job priorities allow a user to sort their HTCondor jobs to determine "
            "which are tried to be run first. "
            "A job priority can be any integer: larger values denote better priority. "
            "By default HTCondor job priority=0. "
        ),
    )
    submission_parser.add(
        "--transfer-files",
        action=StoreBoolean,
        default=True,
        help=(
            "If true (default), use the HTCondor file transfer mechanism"
            " For non-condor schedulers, this option is ignored."
            " Note: the log files are automatically synced, but to sync the "
            " results during the run (e.g. to inspect progress), use the "
            " executable bilby_pipe_htcondor_sync"
        ),
    )
    submission_parser.add(
        "--log-directory",
        type=nonestr,
        default=None,
        help="If given, an alternative path for the log output",
    )
    submission_parser.add(
        "--osg",
        action="store_true",
        default=False,
        help="If true, format condor submission for running on OSG, default is False",
    )
    submission_parser.add(
        "--desired-sites",
        type=nonestr,
        help=(
            "A comma-separated list of desired sites, wrapped in quoates."
            " e.g., desired-sites='site1,site2'. This can be used on the OSG"
            " to specify specific run nodes."
        ),
    )
    submission_parser.add(
        "--analysis-executable",
        default=None,
        type=nonestr,
        help=(
            "Path to an executable to replace bilby_pipe_analysis, be aware"
            " that this executable will pass the complete ini file (in the"
            " outdir.)"
        ),
    )
    submission_parser.add(
        "--analysis-executable-parser",
        default=None,
        type=nonestr,
        help=(
            "Python path to the analysis executable parser, used in conjunction"
            " with analysis-executable. Note, if this is not provided any"
            " new arguments to analysis-executable will raise a warning, but"
            " they will be passed to the executable directly."
        ),
    )

    # likelihood_parser = parser.add_argument_group(
    #     title="Likelihood arguments",
    #     description="Options for setting up the likelihood.",
    # )
    # likelihood_parser.add(
    #     "--distance-marginalization",
    #     action="store_true",
    #     default=False,
    #     help="Boolean. If true, use a distance-marginalized likelihood",
    # )
    # likelihood_parser.add(
    #     "--distance-marginalization-lookup-table",
    #     default=None,
    #     type=nonestr,
    #     help="Path to the distance-marginalization lookup table",
    # )
    #
    # likelihood_parser.add(
    #     "--phase-marginalization",
    #     action="store_true",
    #     default=False,
    #     help="Boolean. If true, use a phase-marginalized likelihood",
    # )
    # likelihood_parser.add(
    #     "--time-marginalization",
    #     action="store_true",
    #     default=False,
    #     help="Boolean. If true, use a time-marginalized likelihood",
    # )
    # likelihood_parser.add(
    #     "--jitter-time",
    #     action=StoreBoolean,
    #     default=True,
    #     help="Boolean. If true, and using a time-marginalized likelihood 'time jittering' will be performed",
    # )
    # likelihood_parser.add(
    #     "--reference-frame",
    #     default="sky",
    #     type=str,
    #     help="Reference frame for the sky parameterisation, either 'sky' (default) or, e.g., 'H1L1'",
    # )
    # likelihood_parser.add(
    #     "--time-reference",
    #     default="geocent",
    #     type=str,
    #     help="Time parameter to sample in, either 'geocent' (default) or, e.g., 'H1'",
    # )
    # likelihood_parser.add(
    #     "--likelihood-type",
    #     default="GravitationalWaveTransient",
    #     help=(
    #         "The likelihood. Can be one of [GravitationalWaveTransient, "
    #         "ROQGravitationalWaveTransient, zero] or python path to a bilby "
    #         "likelihood class available in the users installation. "
    #         "The --roq-folder or both --linear-matrix and --quadratic-matrix "
    #         "are required if the ROQ likelihood used. If both the options are "
    #         "specified, ROQ data are taken from roq-folder, and linear-matrix "
    #         "and quadratic-matrix are ignored."
    #         "If `zero` is given, a testing ZeroLikelihood is used which always"
    #         "return zero."
    #     ),
    # )
    # likelihood_parser.add(
    #     "--roq-folder", type=nonestr, default=None, help="The data for ROQ"
    # )
    # likelihood_parser.add(
    #     "--roq-linear-matrix",
    #     type=nonestr,
    #     default=None,
    #     help="Path to ROQ basis for linear inner products. This option is ignored if roq-folder is not None.",
    # )
    # likelihood_parser.add(
    #     "--roq-quadratic-matrix",
    #     type=nonestr,
    #     default=None,
    #     help="Path to ROQ basis for quadratic inner products. This option is ignored if roq-folder is not None.",
    # )
    # likelihood_parser.add(
    #     "--roq-weights",
    #     type=nonestr,
    #     default=None,
    #     help=(
    #         "If given, the ROQ weights to use (rather than building them). "
    #         "This must be given along with the roq-folder for checking"
    #     ),
    # )
    # likelihood_parser.add(
    #     "--roq-weight-format",
    #     type=nonestr,
    #     default=None,
    #     help=(
    #         "File format of roq weights. This should be npz, hdf5, or json. "
    #         "If not specified, it is set to npz if basis file is specified "
    #         "through roq-folder, and hdf5 if through roq-linear-matrix and "
    #         "roq-quadratic-matrix"
    #     ),
    # )
    # likelihood_parser.add(
    #     "--roq-scale-factor",
    #     default=1,
    #     type=float,
    #     help="Rescaling factor for the ROQ, default is 1 (no rescaling)",
    # )
    # likelihood_parser.add(
    #     "--extra-likelihood-kwargs",
    #     type=nonestr,
    #     default=None,
    #     help="Additional keyword arguments to pass to the likelihood. Any arguments "
    #     "which are named bilby_pipe arguments, e.g., distance_marginalization "
    #     "should NOT be included. This is only used if you are not using the "
    #     "GravitationalWaveTransient or ROQGravitationalWaveTransient likelihoods",
    # )

    output_parser = parser.add_argument_group(
        title="Output arguments", description="What kind of output/summary to generate."
    )
    # output_parser.add_argument(
    #     "--plot-trace",
    #     action="store_true",
    #     help="Create traceplots during the run",
    # )
    output_parser.add_argument(
        "--plot-data",
        action="store_true",
        help="Create plot of the frequency domain data",
    )
    # output_parser.add_argument(
    #     "--plot-injection",
    #     action="store_true",
    #     help="Create time-domain plot of the injection",
    # )
    output_parser.add_argument(
        "--plot-spectrogram",
        action="store_true",
        help="Create spectrogram plot",
    )
    # output_parser.add_argument(
    #     "--plot-calibration",
    #     action="store_true",
    #     help="Create calibration posterior plot",
    # )
    output_parser.add_argument(
        "--plot-corner",
        action="store_true",
        help="Create corner plot",
    )
    output_parser.add_argument(
        "--plot-weights",
        action="store_true",
        help="Create scatter plot of importance weights",
    )
    output_parser.add_argument(
        "--plot-log-probs",
        action="store_true",
        help="Create scatter plot of target versus proposal log probabilities",
    )
    # output_parser.add_argument(
    #     "--plot-marginal",
    #     action="store_true",
    #     help="Create 1-d marginal posterior plots",
    # )
    # output_parser.add_argument(
    #     "--plot-skymap", action="store_true", help="Create posterior skymap"
    # )
    # output_parser.add_argument(
    #     "--plot-waveform", action="store_true", help="Create waveform posterior plot"
    # )
    # output_parser.add_argument(
    #     "--plot-format",
    #     default="png",
    #     help="Format for making bilby_pipe plots, can be [png, pdf, html]. "
    #     "If specified format is not supported, will default to png.",
    # )
    #
    output_parser.add(
        "--create-summary", action="store_true", help="Create a PESummary page"
    )
    output_parser.add("--email", type=nonestr, help="Email for notifications")
    output_parser.add(
        "--notification",
        type=nonestr,
        default="Never",
        help=(
            "Notification setting for HTCondor jobs. "
            "One of 'Always','Complete','Error','Never'. "
            "If defined by 'Always', "
            "the owner will be notified whenever the job "
            "produces a checkpoint, as well as when the job completes. "
            "If defined by 'Complete', "
            "the owner will be notified when the job terminates. "
            "If defined by 'Error', "
            "the owner will only be notified if the job terminates abnormally, "
            "or if the job is placed on hold because of a failure, "
            "and not by user request. "
            "If defined by 'Never' (the default), "
            "the owner will not receive e-mail, regardless to what happens to the job. "
            "Note, an `email` arg is also required for notifications to be emailed. "
        ),
    )
    output_parser.add(
        "--queue",
        type=nonestr,
        help="Condor job queue. Use Online_PE for online parameter estimation runs.",
    )
    output_parser.add(
        "--existing-dir",
        type=nonestr,
        default=None,
        help=(
            "If given, add results to an directory with an an existing"
            " summary.html file"
        ),
    )
    output_parser.add(
        "--webdir",
        type=nonestr,
        default=None,
        help=(
            "Directory to store summary pages. If not given, defaults to "
            "outdir/results_page"
        ),
    )
    output_parser.add(
        "--summarypages-arguments",
        type=nonestr,
        default=None,
        help="Arguments (in the form of a dictionary) to pass to the summarypages executable",
    )
    output_parser.add(
        "--result-format",
        type=str,
        default="hdf5",
        choices=["json", "hdf5", "pickle"],
        help="Format to save the result file in.",
    )
    output_parser.add(
        "--final-result",
        action=StoreBoolean,
        default=True,
        help="If true (default), generate a set of lightweight downsamples final results.",
    )
    output_parser.add(
        "--final-result-nsamples",
        default=20000,
        type=int,
        help="Maximum number of samples to keep in the final results",
    )

    prior_parser = parser.add_argument_group(
        title="Prior arguments", description="Specify the prior settings."
    )
    # prior_parser.add(
    #     "--default-prior",
    #     default="PriorDict",
    #     type=str,
    #     help=(
    #         "The name of the prior set to base the prior on. Can be one of"
    #         "[PriorDict, BBHPriorDict, BNSPriorDict, CalibrationPriorDict]"
    #         "or a python path to a bilby prior class available in the user's installation."
    #     ),
    # )
    # prior_parser.add(
    #     "--deltaT",
    #     type=float,
    #     default=0.2,
    #     help=(
    #         "The symmetric width (in s) around the trigger time to"
    #         " search over the coalescence time"
    #     ),
    # )
    prior_parser_main = prior_parser.add_mutually_exclusive_group()
    # prior_parser_main.add(
    #     "--prior-file", type=nonestr, default=None, help="The prior file"
    # )
    prior_parser_main.add(
        "--prior-dict",
        type=nonestr,
        default=None,
        help=(
            "A dictionary of priors (alternative to prior-file). Multiline "
            "dictionaries are supported, but each line must contain a single"
            "parameter specification and finish with a comma. Dingo priors are set at "
            "network training time, so the prior-dict is used at importance sampling "
            "time to re-weight to the new prior. The prior-dict does not have to be "
            "a prior over the entire set of parameters, only the parameters for which "
            "the prior is changed."
        ),
    )
    # prior_parser.add(
    #     "--enforce-signal-duration",
    #     action=StoreBoolean,
    #     default=True,
    #     help=(
    #         "Whether to require that all signals fit within the segment duration. "
    #         "The signal duration is calculated using a post-Newtonian approximation."
    #     ),
    # )

    # postprocessing_parser = parser.add_argument_group(
    #     title="Post processing arguments",
    #     description="What post-processing to perform.",
    # )
    # postprocessing_parser.add(
    #     "--postprocessing-executable",
    #     type=nonestr,
    #     default=None,
    #     help=(
    #         "An executable name for postprocessing. A single postprocessing "
    #         " job is run as a child of all analysis jobs"
    #     ),
    # )
    # postprocessing_parser.add(
    #     "--postprocessing-arguments",
    #     type=nonestr,
    #     default=None,
    #     help="Arguments to pass to the postprocessing executable",
    # )
    # postprocessing_parser.add(
    #     "--single-postprocessing-executable",
    #     type=nonestr,
    #     default=None,
    #     help=(
    #         "An executable name for postprocessing. A single postprocessing "
    #         "job is run as a child for each analysis jobs: note the "
    #         "difference with respect postprocessing-executable"
    #     ),
    # )
    # postprocessing_parser.add(
    #     "--single-postprocessing-arguments",
    #     type=nonestr,
    #     default=None,
    #     help=(
    #         "Arguments to pass to the single postprocessing executable. The "
    #         "str '$RESULT' will be replaced by the path to the individual "
    #         "result file"
    #     ),
    # )

    # sampler_parser = parser.add_argument_group(title="Sampler arguments")
    # sampler_parser.add("--sampler", type=str, default="dynesty", help="Sampler to use")
    # sampler_parser.add(
    #     "--sampling-seed", default=None, type=noneint, help="Random sampling seed"
    # )

    # sampler_parser.add(
    #     "--sampler-kwargs",
    #     type=str,
    #     default="DynestyDefault",
    #     help=(
    #         "Dictionary of sampler-kwargs to pass in, e.g., {nlive: 1000} OR "
    #         "pass pre-defined set of sampler-kwargs {DynestyDefault, BilbyMCMCDefault, FastTest}"
    #     ),
    # )

    # Waveform arguments
    waveform_parser = parser.add_argument_group(
        title="Waveform arguments", description="Setting for the waveform generator"
    )
    # waveform_parser.add(
    #     "--waveform-generator",
    #     default="bilby.gw.waveform_generator.LALCBCWaveformGenerator",
    #     type=str,
    #     help="The waveform generator class, should be a python path. This will "
    #     "not be able to use any arguments not passed to the default.",
    # )
    waveform_parser.add(
        "--reference-frequency",
        default=None,
        type=nonefloat,
        help="The reference " "frequency",
    )
    waveform_parser.add(
        "--waveform-approximant",
        default=None,
        type=nonestr,
        help="The name of the waveform approximant to use for PE.",
    )
    # waveform_parser.add(
    #     "--catch-waveform-errors",
    #     default=True,
    #     action=StoreBoolean,
    #     help="Turns on waveform error catching",
    # )
    # waveform_parser.add(
    #     "--pn-spin-order",
    #     default=-1,
    #     type=int,
    #     help="Post-newtonian order to use for the spin",
    # )
    # waveform_parser.add(
    #     "--pn-tidal-order",
    #     default=-1,
    #     type=int,
    #     help="Post-Newtonian order to use for tides",
    # )
    # waveform_parser.add(
    #     "--pn-phase-order",
    #     default=-1,
    #     type=int,
    #     help="post-Newtonian order to use for the phase",
    # )
    # waveform_parser.add(
    #     "--pn-amplitude-order",
    #     default=0,
    #     type=int,
    #     help="Post-Newtonian order to use for the amplitude. Also "
    #     "used to determine the waveform starting frequency.",
    # )
    # waveform_parser.add(
    #     "--numerical-relativity-file",
    #     default=None,
    #     type=nonestr,
    #     help=(
    #         "Path to a h5 numerical relativity file to inject, see"
    #         "https://git.ligo.org/waveforms/lvcnr-lfs for examples"
    #     ),
    # )
    # waveform_parser.add(
    #     "--waveform-arguments-dict",
    #     default=None,
    #     type=nonestr,
    #     help=(
    #         "A dictionary of arbitrary additional waveform-arguments to pass"
    #         "  to the bilby waveform generator's `waveform_arguments`"
    #     ),
    # )
    # waveform_parser.add(
    #     "--mode-array",
    #     default=None,
    #     type=nonestr,
    #     action="append",
    #     help="Array of modes to use for the waveform. Should be "
    #     "a list of lists, eg. [[2,2], [2,-2]]",
    # )
    # waveform_parser.add(
    #     "--frequency-domain-source-model",
    #     default="lal_binary_black_hole",
    #     type=str,
    #     help=(
    #         "Name of the frequency domain source model. Can be one of"
    #         "[lal_binary_black_hole, lal_binary_neutron_star,"
    #         "lal_eccentric_binary_black_hole_no_spins, sinegaussian, "
    #         "supernova, supernova_pca_model] or any python  path to a bilby "
    #         " source function the users installation, e.g. examp.source.bbh"
    #     ),
    # )
    # waveform_parser.add(
    #     "--conversion-function",
    #     default=None,
    #     type=nonestr,
    #     help=(
    #         "Optional python path to a user-specified conversion function "
    #         "If unspecified, this is determined by the frequency_domain_source_model."
    #         "If the source-model contains binary_black_hole, "
    #         "the conversion function is bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters. "
    #         "If the source-model contains binary_neutron_star, "
    #         "the generation function is bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters. "
    #         "If you specify your own function, you may wish to use the I/O of those functions as templates."
    #         "If given as 'noconvert' (case insensitive), no conversion is used'"
    #     ),
    # )
    # waveform_parser.add(
    #     "--generation-function",
    #     default=None,
    #     type=nonestr,
    #     help=(
    #         "Optional python path to a user-specified generation function "
    #         "If unspecified, this is determined by the frequency_domain_source_model."
    #         "If the source-model contains binary_black_hole, "
    #         "the generation function is bilby.gw.conversion.generate_all_bbh_parameters. "
    #         "If the source-model contains binary_neutron_star, "
    #         "the generation function is bilby.gw.conversion.generate_all_bns_parameters. "
    #         "If you specify your own function, you may wish to use the I/O of those functions as templates"
    #         "If given as 'noconvert' (case insensitive), no generation is used'"
    #     ),
    # )

    #
    # Added for Dingo.
    #

    sampler_parser = parser.add_argument_group(title="Sampler arguments")
    sampler_parser.add(
        "--model",
        type=str,
        required=True,
        help="Neural network model for generating posterior samples.",
    )
    sampler_parser.add(
        "--model-init",
        type=nonestr,
        default=None,
        help="Neural network model for generating samples to initialize Gibbs sampling."
        "Must be provided if the main model is a GNPE model. ",
    )
    sampler_parser.add(
        "--recover-log-prob",
        action=StoreBoolean,
        default=True,
        help="For GNPE models, whether to recover the log probability by training an "
        "unconditional flow for the GNPE proxies.",
    )
    sampler_parser.add(
        "--device",
        type=str,
        default="cuda",
        choices=["cuda", "cpu"],
        help="Device to use for sampling. Choices are 'cpu' and 'cuda'. Default 'cuda'.",
    )
    sampler_parser.add(
        "--num-gnpe-iterations",
        type=int,
        default=30,
        help="Number of GNPE iterations to perform when using a GNPE model. Default 30.",
    )
    sampler_parser.add(
        "--num-samples",
        type=int,
        default=50000,
        help="Number of posterior samples desired. Default is 50000",
    )
    sampler_parser.add(
        "--batch-size",
        type=int,
        default=50000,
        help="Number of samples per batch. This is limited by the amount of GPU RAM. "
        "Default is 50000",
    )
    sampler_parser.add(
        "--density-recovery-settings",
        type=str,
        default="ProxyRecoveryDefault",
        help=(
            "Dictionary of density-recovery-settings to pass in, e.g., {num_samples: "
            "400_000, nde_settings: (...)} OR pass pre-defined set of "
            "density-recovery-settings {ProxyRecoveryDefault}"
        ),
    )
    sampler_parser.add(
        "--importance-sample",
        action=StoreBoolean,
        default=True,
        help=(
            "Whether to perform importance sampling on result. (Default: True)"
        ),
    )
    sampler_parser.add(
        "--importance-sampling-settings",
        type=str,
        default="Default",
        help=(
            "Dictionary of importance-sampling-settings to pass in, e.g., "
            "{synthetic_phase: {approximation_22_mode: False, (...)}} OR pass"
            "pre-defined set of density-recovery-settings {PhaseRecoveryDefault}"
        ),
    )
    sampler_parser.add(
        "--importance-sampling-updates",
        type=str,
        default="{}",
        help=(
            "Dictionary of updated settings to be used for importance sampling, "
            "including new prior, data conditioning, and waveform approximant. This "
            "will get populated with any settings provided elsewhere that would "
            "otherwise override model settings. This is useful for tweaking settings "
            "that networks were trained with."
        ),
    )
    sampler_parser.add(
        "--n-parallel",
        type=int,
        default=1,
        help=(
            "This sets the number of parallel condor jobs for importance sampling. Each"
            "of these jobs can in turn be parallelized across several CPU cores. The "
            "total number of processes is n-parallel * "
            "request-cpus-importance-sampling."
        ),
    )

    return parser



import ast
import h5py
import numpy as np
import pandas as pd

from dingo.core.utils.misc import get_version


def recursive_hdf5_save(group, d):
    for k, v in d.items():
        if v is None:
            continue
        elif isinstance(v, dict):
            next_group = group.create_group(k)
            recursive_hdf5_save(next_group, v)
        elif isinstance(v, np.ndarray):
            group.create_dataset(k, data=v)
        elif isinstance(v, pd.DataFrame):
            group.create_dataset(k, data=v.to_records(index=False))
        elif isinstance(v, (int, float, str, list)):
            # TODO: Set scalars as attributes?
            group.create_dataset(k, data=v)
        else:
            raise TypeError(f"Cannot save datatype {type(v)} as hdf5 dataset.")


def recursive_hdf5_load(group, keys=None):
    d = {}
    for k, v in group.items():
        if keys is None or k in keys:
            if isinstance(v, h5py.Group):
                d[k] = recursive_hdf5_load(v)
            else:
                d[k] = v[...]
                # If the array has column names, load it as a pandas DataFrame
                if d[k].dtype.names is not None:
                    d[k] = pd.DataFrame(d[k])
                # Convert arrays of size 1 to scalars
                elif d[k].size == 1:
                    d[k] = d[k].item()
                    if isinstance(d[k], bytes):
                        # Assume this is a string.
                        d[k] = d[k].decode()
                # If an array is 1D and of type object, assume it originated as a list
                # of strings.
                elif d[k].ndim == 1 and d[k].dtype == "O":
                    d[k] = [x.decode() for x in d[k]]
    return d


class DingoDataset:
    """This is a generic dataset class with save / load methods.

    A common use case is to inherit multiply from DingoDataset and
    torch.utils.data.Dataset, in which case the subclass picks up these I/O methods,
    and DingoDataset is acting as a Mixin class.

    Alternatively, if the torch Dataset is not needed, then DingoDataset can be
    subclassed directly.
    """

    dataset_type = "dingo_dataset"

    def __init__(self, file_name=None, dictionary=None, data_keys=None):
        """
        For constructing, provide either file_name, or dictionary containing data and
        settings entries, or neither.

        Parameters
        ----------
        file_name : str
            HDF5 file containing a dataset
        dictionary : dict
            Contains settings and data entries. The data keys should be the same as
            save_keys
        data_keys : list
            Variables that should be saved / loaded. This allows for class to store
            additional variables beyond those that are saved. Typically, this list
            would be provided by any subclass.
        """
        self._data_keys = list(data_keys)  # Make a copy before modifying.
        self._data_keys.append("version")

        # Ensure all potential variables have None values to begin
        for key in self._data_keys:
            vars(self)[key] = None
        self.settings = None
        self.version = None

        # If data provided, load it
        if file_name is not None:
            self.from_file(file_name)
        elif dictionary is not None:
            self.from_dictionary(dictionary)

    def to_file(self, file_name, mode="w"):
        print("Saving dataset to " + str(file_name))
        save_dict = {
            k: v
            for k, v in vars(self).items()
            if k in self._data_keys and v is not None
        }
        with h5py.File(file_name, mode) as f:
            recursive_hdf5_save(f, save_dict)
            if self.settings:
                f.attrs["settings"] = str(self.settings)
            if self.dataset_type:
                f.attrs["dataset_type"] = self.dataset_type

    def from_file(self, file_name):
        print("Loading dataset from " + str(file_name) + ".")
        with h5py.File(file_name, "r") as f:
            # Load only the keys that the class expects
            loaded_dict = recursive_hdf5_load(f, keys=self._data_keys)
            for k, v in loaded_dict.items():
                assert k in self._data_keys
                vars(self)[k] = v
            try:
                self.settings = ast.literal_eval(f.attrs["settings"])
            except KeyError:
                self.settings = None  # Is this necessary?

    def to_dictionary(self):
        dictionary = {
            k: v
            for k, v in vars(self).items()
            if (k in self._data_keys or k == "settings") and v is not None
        }
        return dictionary

    def from_dictionary(self, dictionary: dict):
        for k, v in dictionary.items():
            if k in self._data_keys or k == "settings":
                vars(self)[k] = v
        if "version" not in dictionary:
            self.version = f"dingo={get_version()}"



from copy import deepcopy

from bilby.gw.prior import BBHPriorDict
from bilby.gw.conversion import (
    fill_from_fixed_priors,
    convert_to_lal_binary_black_hole_parameters,
)
from bilby.core.prior import Uniform, Sine, Cosine

import numpy as np
from typing import Dict, Set, Any
import warnings

# Silence INFO and WARNING messages from bilby
import logging

logging.getLogger("bilby").setLevel("ERROR")


class BBHExtrinsicPriorDict(BBHPriorDict):
    """
    This class is the same as BBHPriorDict except that it does not require mass parameters.

    It also contains a method for estimating the standardization parameters.

    TODO:
        * Add support for zenith/azimuth
        * Defaults?
    """

    def default_conversion_function(self, sample):
        out_sample = fill_from_fixed_priors(sample, self)
        out_sample, _ = convert_to_lal_binary_black_hole_parameters(out_sample)

        # The previous call sometimes adds phi_jl, phi_12 parameters. These are
        # not needed so they can be deleted.
        if "phi_jl" in out_sample.keys():
            del out_sample["phi_jl"]
        if "phi_12" in out_sample.keys():
            del out_sample["phi_12"]

        return out_sample

    def mean_std(self, keys=([]), sample_size=50000, force_numerical=False):
        """
        Calculate the mean and standard deviation over the prior.

        Parameters
        ----------
        keys: list(str)
            A list of desired parameter names
        sample_size: int
            For nonanalytic priors, number of samples to use to estimate the
            result.
        force_numerical: bool (False)
            Whether to force a numerical estimation of result, even when
            analytic results are available (useful for testing)

        Returns dictionaries for the means and standard deviations.

        TODO: Fix for constrained priors. Shouldn't be an issue for extrinsic parameters.
        """
        mean = {}
        std = {}

        if not force_numerical:
            # First try to calculate analytically (works for standard priors)
            estimation_keys = []
            for key in keys:
                p = self[key]
                # A few analytic cases
                if isinstance(p, Uniform):
                    mean[key] = (p.maximum + p.minimum) / 2.0
                    std[key] = np.sqrt((p.maximum - p.minimum) ** 2.0 / 12.0).item()
                elif isinstance(p, Sine) and p.minimum == 0.0 and p.maximum == np.pi:
                    mean[key] = np.pi / 2.0
                    std[key] = np.sqrt(0.25 * (np.pi ** 2) - 2).item()
                elif (
                    isinstance(p, Cosine)
                    and p.minimum == -np.pi / 2
                    and p.maximum == np.pi / 2
                ):
                    mean[key] = 0.0
                    std[key] = np.sqrt(0.25 * (np.pi ** 2) - 2).item()
                else:
                    estimation_keys.append(key)
        else:
            estimation_keys = keys

        # For remaining parameters, estimate numerically
        if len(estimation_keys) > 0:
            samples = self.sample_subset(keys, size=sample_size)
            samples = self.default_conversion_function(samples)
            for key in estimation_keys:
                if key in samples.keys():
                    mean[key] = np.mean(samples[key]).item()
                    std[key] = np.std(samples[key]).item()

        return mean, std


# TODO: Add latex labels, names


default_extrinsic_dict = {
    "dec": "bilby.core.prior.Cosine(minimum=-np.pi/2, maximum=np.pi/2)",
    "ra": 'bilby.core.prior.Uniform(minimum=0., maximum=2*np.pi, boundary="periodic")',
    "geocent_time": "bilby.core.prior.Uniform(minimum=-0.1, maximum=0.1)",
    "psi": 'bilby.core.prior.Uniform(minimum=0.0, maximum=np.pi, boundary="periodic")',
    "luminosity_distance": "bilby.core.prior.Uniform(minimum=100.0, maximum=6000.0)",
}

default_intrinsic_dict = {
    "mass_1": "bilby.core.prior.Constraint(minimum=10.0, maximum=80.0)",
    "mass_2": "bilby.core.prior.Constraint(minimum=10.0, maximum=80.0)",
    "mass_ratio": "bilby.gw.prior.UniformInComponentsMassRatio(minimum=0.125, maximum=1.0)",
    "chirp_mass": "bilby.gw.prior.UniformInComponentsChirpMass(minimum=25.0, maximum=100.0)",
    "luminosity_distance": 1000.0,
    "theta_jn": "bilby.core.prior.Sine(minimum=0.0, maximum=np.pi)",
    "phase": 'bilby.core.prior.Uniform(minimum=0.0, maximum=2*np.pi, boundary="periodic")',
    "a_1": "bilby.core.prior.Uniform(minimum=0.0, maximum=0.99)",
    "a_2": "bilby.core.prior.Uniform(minimum=0.0, maximum=0.99)",
    "tilt_1": "bilby.core.prior.Sine(minimum=0.0, maximum=np.pi)",
    "tilt_2": "bilby.core.prior.Sine(minimum=0.0, maximum=np.pi)",
    "phi_12": 'bilby.core.prior.Uniform(minimum=0.0, maximum=2*np.pi, boundary="periodic")',
    "phi_jl": 'bilby.core.prior.Uniform(minimum=0.0, maximum=2*np.pi, boundary="periodic")',
    "geocent_time": 0.0,
}

default_inference_parameters = [
    "chirp_mass",
    "mass_ratio",
    "phase",
    "a_1",
    "a_2",
    "tilt_1",
    "tilt_2",
    "phi_12",
    "phi_jl",
    "theta_jn",
    "luminosity_distance",
    "geocent_time",
    "ra",
    "dec",
    "psi",
]


def build_prior_with_defaults(prior_settings: Dict[str, str]):
    """
    Generate BBHPriorDict based on dictionary of prior settings,
    allowing for default values.

    Parameters
    ----------
    prior_settings: Dict
        A dictionary containing prior definitions for intrinsic parameters
        Allowed values for each parameter are:
            * 'default' to use a default prior
            * a string for a custom prior, e.g.,
               "Uniform(minimum=10.0, maximum=80.0, name=None, latex_label=None, unit=None, boundary=None)"

    Depending on the particular prior choices the dimensionality of a
    parameter sample obtained from the returned GWPriorDict will vary.
    """

    full_prior_settings = deepcopy(prior_settings)
    for k, v in prior_settings.items():
        if v == "default":
            full_prior_settings[k] = default_intrinsic_dict[k]

    return BBHPriorDict(full_prior_settings)


def split_off_extrinsic_parameters(theta):
    """
    Split theta into intrinsic and extrinsic parameters.

    Parameters
    ----------
    theta: dict
        BBH parameters. Includes intrinsic parameters to be passed to waveform
        generator, and extrinsic parameters for detector projection.

    Returns
    -------
    theta_intrinsic: dict
        BBH intrinsic parameters.
    theta_extrinsic: dict
        BBH extrinsic parameters.
    """
    extrinsic_parameters = ["geocent_time", "luminosity_distance", "ra", "dec", "psi"]
    theta_intrinsic = {}
    theta_extrinsic = {}
    for k, v in theta.items():
        if k in extrinsic_parameters:
            theta_extrinsic[k] = v
        else:
            theta_intrinsic[k] = v
    # set fiducial values for time and distance
    theta_intrinsic["geocent_time"] = 0
    theta_intrinsic["luminosity_distance"] = 100
    return theta_intrinsic, theta_extrinsic



"""
This tests the method WaveformGenerator.generate_hplus_hcross_m, that returns the
polarzations disentangled into contributions m \in [-l_max, ...,0, ...,l_max],
that transform as exp(-1j * m * phase_shift) under phase shifts. This is important when
treating the phase parameter as an extrinsic parameter.

Note: this only accounts for the modified argument in the spherical harmonics, not for
the rotation of phase_shift of the cartesian spins in xy plane. Our workaround is to
set wfg.spin_conversion_phase = 0.0, which sets a constant phase 0 when converting PE
spins to cartesian spins. This means that phi_12 and phi_jl have different definitions,
which needs to be accounted for in postprocessing. The tests below all use
wfg.spin_conversion_phase = 0.0.
"""
import pytest
import numpy as np

from dingo.gw.waveform_generator import WaveformGenerator, sum_contributions_m
from dingo.gw.gwutils import get_mismatch
from dingo.gw.domains import build_domain
from dingo.gw.prior import build_prior_with_defaults


@pytest.fixture
def uniform_fd_domain():
    domain_settings = {
        "type": "FrequencyDomain",
        "f_min": 10.0,
        "f_max": 2048.0,
        "delta_f": 0.125,
    }
    domain = build_domain(domain_settings)
    return domain


@pytest.fixture
def intrinsic_prior():
    intrinsic_dict = {
        "mass_1": "bilby.core.prior.Constraint(minimum=10.0, maximum=80.0)",
        "mass_2": "bilby.core.prior.Constraint(minimum=10.0, maximum=80.0)",
        "mass_ratio": "bilby.gw.prior.UniformInComponentsMassRatio(minimum=0.125, maximum=1.0)",
        "chirp_mass": "bilby.gw.prior.UniformInComponentsChirpMass(minimum=25.0, maximum=100.0)",
        "luminosity_distance": 1000.0,
        "theta_jn": "bilby.core.prior.Sine(minimum=0.0, maximum=np.pi)",
        "phase": 'bilby.core.prior.Uniform(minimum=0.0, maximum=2*np.pi, boundary="periodic")',
        "a_1": "bilby.core.prior.Uniform(minimum=0.0, maximum=0.99)",
        "a_2": "bilby.core.prior.Uniform(minimum=0.0, maximum=0.99)",
        "tilt_1": "bilby.core.prior.Sine(minimum=0.0, maximum=np.pi)",
        "tilt_2": "bilby.core.prior.Sine(minimum=0.0, maximum=np.pi)",
        "phi_12": 'bilby.core.prior.Uniform(minimum=0.0, maximum=2*np.pi, boundary="periodic")',
        "phi_jl": 'bilby.core.prior.Uniform(minimum=0.0, maximum=2*np.pi, boundary="periodic")',
        "geocent_time": 0.0,
    }
    prior = build_prior_with_defaults(intrinsic_dict)
    return prior


def test_generate_hplus_hcross_m_SEOBNRv4PHM(uniform_fd_domain, intrinsic_prior):
    domain = uniform_fd_domain
    prior = intrinsic_prior
    wfg = WaveformGenerator(
        "SEOBNRv4PHM",
        domain,
        10.0,
        f_start=10.0,
        spin_conversion_phase=0.0,
    )

    p = prior.sample()
    phase_shift = np.random.uniform(high=2 * np.pi)
    pol_m = wfg.generate_hplus_hcross_m(p)
    pol = sum_contributions_m(pol_m, phase_shift=phase_shift)
    pol_ref = wfg.generate_hplus_hcross({**p, "phase": p["phase"] + phase_shift})

    mismatches = [
        get_mismatch(
            pol[pol_name],
            pol_ref[pol_name],
            wfg.domain,
            asd_file="aLIGO_ZERO_DET_high_P_asd.txt",
        )
        for pol_name in pol
    ]

    # The mismatches are typically be of order 1e-5. This is exclusively due to
    # different tapering. The reference polarizations are tapered and FFTed on the
    # level of polarizations, while for generate_hplus_hcross_m, the tapering and FFT
    # happens on the level of complex modes.
    # We tested the mismatches for 20k waveforms, and the largest mismatch encountered
    # was 7e-4, while almost all mismatches were of order 1e-5.
    assert max(mismatches) < 5e-4


def test_generate_hplus_hcross_m_IMRPhenomXPHM(uniform_fd_domain, intrinsic_prior):
    domain = uniform_fd_domain
    prior = intrinsic_prior
    wfg = WaveformGenerator(
        "IMRPhenomXPHM",
        domain,
        20.0,
        f_start=10.0,
        spin_conversion_phase=0.0,
    )

    mismatches = []
    for idx in range(10):
        p = prior.sample()
        phase_shift = np.random.uniform(high=2 * np.pi)

        pol_m = wfg.generate_hplus_hcross_m(p)
        pol = sum_contributions_m(pol_m, phase_shift=phase_shift)
        pol_ref = wfg.generate_hplus_hcross({**p, "phase": p["phase"] + phase_shift})

        mismatches.append(
            [
                get_mismatch(
                    pol[pol_name],
                    pol_ref[pol_name],
                    wfg.domain,
                    asd_file="aLIGO_ZERO_DET_high_P_asd.txt",
                )
                for pol_name in pol
            ]
        )

    # The mismatches are typically be of order 1e-5 to 1e-9. This comes from the
    # calculation of the magnitude of the orbital angular momentum, which we calculate
    # to a different order the IMRPhenomXPHM. It's tricky to get this exactly right,
    # since there are many different methods for this. But the small mismatches we do
    # get should not have a big effect in practice.
    mismatches = np.array(mismatches)
    assert np.max(mismatches) < 2e-3



# adapted from the Asimov Bilby Pipeline interface
import configparser
import glob
import importlib
import os
import re
import subprocess
import time


from asimov import config, logger

from asimov.pipeline import Pipeline, PipelineException, PipelineLogger, PESummaryPipeline


class Dingo(Pipeline):
    """
    The Dingo Pipeline.

    Parameters
    ----------
    production : :class:`asimov.Production`
       The production object.
    category : str, optional
        The category of the job.
        Defaults to "C01_offline".
    """

    # config_template = importlib.resources.path(__name__, 'dingo.ini')
    name = "dingo"
    STATUS = {"wait", "stuck", "stopped", "running", "finished"}

    def __init__(self, production, category=None):
        super(Dingo, self).__init__(production, category)
        self.logger = logger
        if not production.pipeline.lower() == self.name:
            raise PipelineException

    def detect_completion(self):
        """
        Check for the production of the posterior file to signal that the job has completed.
        """
        self.logger.info("Checking if the dingo job has completed")

        results_dir = glob.glob(f"{self.production.rundir}/result")
        if len(results_dir) > 0:
            if (
                len(glob.glob(os.path.join(results_dir[0], f"*importance_sampling.hdf5")))
                > 0
            ):
                self.logger.info("Results files found, the job is finished.")
                return True
            else:
                self.logger.info("No results files found.")
                return False
        else:
            self.logger.info("No results directory found")
            return False

    def before_submit(self):
        """Pre-submit hook."""
        self.logger.info("Running the before_submit hook")
        pass

    def build_dag(self, psds=None, user=None, clobber_psd=False, dryrun=False):
        """
        Construct a DAG file in order to submit a production to the
        condor scheduler using DingoPipe.

        Parameters
        ----------
        production : str
           The production name.
        psds : dict, optional
           The PSDs which should be used for this DAG. If no PSDs are
           provided the PSD files specified in the ini file will be used
           instead.
        user : str
           The user accounting tag which should be used to run the job.
        dryrun: bool
           If set to true the commands will not be run, but will be printed to standard output. Defaults to False.



        Raises
        ------
        PipelineException
           Raised if the construction of the DAG fails.
        """

        cwd = os.getcwd()
        self.logger.info(f"Working in {cwd}")

        if self.production.event.repository:
            ini = self.production.event.repository.find_prods(
                self.production.name, self.category
            )[0]
            ini = os.path.join(cwd, ini)
        else:
            ini = f"{self.production.name}.ini"

        if self.production.rundir:
            rundir = self.production.rundir
        else:
            rundir = os.path.join(
                os.path.expanduser("~"),
                self.production.event.name,
                self.production.name,
            )
            self.production.rundir = rundir

        if "job label" in self.production.meta:
            job_label = self.production.meta["job label"]
        else:
            job_label = self.production.name

        command = [
            os.path.join(config.get("pipelines", "environment"), "bin", "dingo_pipe"),
            ini,
            "--label",
            job_label,
            "--outdir",
            f"{cwd}/{self.production.rundir}",
            "--accounting",
            f"{self.production.meta['scheduler']['accounting group']}",
        ]

        if dryrun:
            print(" ".join(command))
        else:
            self.logger.info(" ".join(command))
            pipe = subprocess.Popen(
                command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )
            out, err = pipe.communicate()
            self.logger.info(out)

            if err or "DAG generation complete, to submit jobs" not in str(out):
                self.logger.error(err)
                raise PipelineException(
                    f"DAG file could not be created.\n{command}\n{out}\n\n{err}",
                    production=self.production.name,
                )

            else:
                time.sleep(10)
                return PipelineLogger(message=out, production=self.production.name)

    def samples(self):
        """
        Collect the combined samples files for PESummary.
        """

        results_dir = os.path.join(self.production.rundir, "result")
        results_filenames = glob.glob(os.path.join(results_dir, f"*importance_sampling.hdf5"))
        return os.path.join(results_dir, results_filenames[0])

    def upload_assets(self):
        """
        Upload the samples from this job.
        """

        asset = self.collect_assets()["samples"]
        self.production.event.repository.add_file(
            asset,
            "posterior_samples.hdf5",
            commit_message=f"Added posterior_samples.",
        )

    def collect_assets(self):
        """
        Gather all of the results assets for this job.
        """
        return {"samples": self.samples()}

    def collect_logs(self):
        """
        Collect all of the log files which have been produced by this production and
        return their contents as a dictionary.
        """
        logs = glob.glob(f"{self.production.rundir}/submit/*.err") + glob.glob(
            f"{self.production.rundir}/log*/*.err"
        )
        logs += glob.glob(f"{self.production.rundir}/*/*.out")
        messages = {}
        for log in logs:
            try:
                with open(log, "r") as log_f:
                    message = log_f.read()
                    message = message.split("\n")
                    messages[log.split("/")[-1]] = "\n".join(message[-100:])
            except FileNotFoundError:
                messages[
                    log.split("/")[-1]
                ] = "There was a problem opening this log file."
        return messages

    def submit_dag(self, dryrun=False):
        """
        Submit a DAG file to the condor cluster.

        Parameters
        ----------
        dryrun : bool
           If set to true the DAG will not be submitted,
           but all commands will be printed to standard
           output instead. Defaults to False.

        Returns
        -------
        int
           The cluster ID assigned to the running DAG file.
        PipelineLogger
           The pipeline logger message.

        Raises
        ------
        PipelineException
           This will be raised if the pipeline fails to submit the job.
        """

        cwd = os.getcwd()
        self.logger.info(f"Working in {cwd}")

        self.before_submit()

        try:
            if "job label" in self.production.meta:
                job_label = self.production.meta["job label"]
            else:
                job_label = self.production.name
            dag_filename = f"dag_{job_label}.submit"
            command = [
                "condor_submit_dag",
                "-batch-name",
                f"dingo/{self.production.event.name}/{self.production.name}",
                os.path.join(self.production.rundir, "submit", dag_filename),
            ]

            if dryrun:
                print(" ".join(command))
            else:
                dagman = subprocess.Popen(
                    command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                )

                self.logger.info(" ".join(command))

                stdout, stderr = dagman.communicate()

                if "submitted to cluster" in str(stdout):
                    cluster = re.search(
                        r"submitted to cluster ([\d]+)", str(stdout)
                    ).groups()[0]
                    self.logger.info(
                        f"Submitted successfully. Running with job id {int(cluster)}"
                    )
                    self.production.status = "running"
                    self.production.job_id = int(cluster)
                    return cluster, PipelineLogger(stdout)
                else:
                    self.logger.error("Could not submit the job to the cluster")
                    self.logger.info(stdout)
                    self.logger.error(stderr)

                    raise PipelineException(
                        "The DAG file could not be submitted.",
                    )

        except FileNotFoundError as error:
            self.logger.exception(error)
            raise PipelineException(
                "It looks like condor isn't installed on this system.\n"
                f"""I wanted to run {" ".join(command)}."""
            ) from error

    def after_completion(self):
        post_pipeline = PESummaryPipeline(production=self.production)
        self.logger.info("Job has completed. Running PE Summary.")
        cluster = post_pipeline.submit_dag()
        self.production.meta["job id"] = int(cluster)
        self.production.status = "processing"
        self.production.event.update_data()

    def detect_completion_processing(self):
        # no post processing currently performed
        return True

    def resurrect(self):
        """
        Attempt to ressurrect a failed job.
        """
        try:
            count = self.production.meta["resurrections"]
        except:
            count = 0
        if (count < 5) and (
            len(glob.glob(os.path.join(self.production.rundir, "submit", "*.rescue*")))
            > 0
        ):
            count += 1
            self.submit_dag()

    @classmethod
    def read_ini(cls, filepath):
        """
        Read and parse a dingo configuration file.

        Parameters
        ----------
        filepath: str
           The path to the ini file.
        """

        with open(filepath, "r") as f:
            file_content = f.read()

        config_parser = configparser.RawConfigParser()
        config_parser.read_string(file_content)

        return config_parser



import copy

import pandas as pd
import torch.multiprocessing
import torchvision
from threadpoolctl import threadpool_limits
from torch.utils.data import DataLoader
from bilby.gw.detector import InterferometerList

from dingo.gw.SVD import SVDBasis

from dingo.gw.dataset.waveform_dataset import WaveformDataset
from dingo.gw.domains import build_domain
from dingo.gw.transforms import (
    ProjectOntoDetectors,
    SampleNoiseASD,
    WhitenAndScaleStrain,
    AddWhiteNoiseComplex,
    SelectStandardizeRepackageParameters,
    RepackageStrainsAndASDS,
    UnpackDict,
    GNPECoalescenceTimes,
    SampleExtrinsicParameters,
    GetDetectorTimes,
)
from dingo.gw.noise.asd_dataset import ASDDataset
from dingo.gw.prior import default_inference_parameters
from dingo.gw.gwutils import *
from dingo.core.utils import *


def build_dataset(data_settings):
    """Build a dataset based on a settings dictionary. This should contain the path of
    a saved waveform dataset.

    This function also truncates the dataset as necessary.

    Parameters
    ----------
    data_settings : dict

    Returns
    -------
    WaveformDataset
    """
    # Build and truncate datasets
    domain_update = data_settings.get("domain_update", None)
    wfd = WaveformDataset(
        file_name=data_settings["waveform_dataset_path"],
        precision="single",
        domain_update=domain_update,
        svd_size_update=data_settings.get("svd_size_update"),
    )
    return wfd


def set_train_transforms(wfd, data_settings, asd_dataset_path, omit_transforms=None):
    """
    Set the transform attribute of a waveform dataset based on a settings dictionary.
    The transform takes waveform polarizations, samples random extrinsic parameters,
    projects to detectors, adds noise, and formats the data for input to the neural
    network. It also implements optional GNPE transformations.

    Note that the WaveformDataset is modified in-place, so this function returns nothing.

    Parameters
    ----------
    wfd : WaveformDataset
    data_settings : dict
    asd_dataset_path : str
        Path corresponding to the ASD dataset used to generate noise.
    omit_transforms :
        List of sub-transforms to omit from the full composition.
    """

    print(f"Setting train transforms.")
    if omit_transforms is not None:
        print("Omitting \n\t" + "\n\t".join([t.__name__ for t in omit_transforms]))

    # By passing the wfd domain when instantiating the noise dataset, this ensures the
    # domains will match. In particular, it truncates the ASD dataset beyond the new
    # f_max, and sets it to 1 below f_min.
    asd_dataset = ASDDataset(
        asd_dataset_path,
        ifos=data_settings["detectors"],
        precision="single",
        domain_update=wfd.domain.domain_dict,
    )
    assert wfd.domain == asd_dataset.domain

    # Add window factor to domain, so that we can compute the noise variance.
    # TODO: we want to set `domain = wfd.domain`. This does not work at the moment,
    #  because this requires updating the window factor of the wfd.domain (instead of
    #  just the local domain object). This causes trouble if the
    #  set_train_transforms function is called multiple times, since the second time
    #  the domain_update = wfd.domain.domain_dict contains a window factor, which will
    #  cause an error in domain_update.
    domain = build_domain(wfd.domain.domain_dict)
    domain.window_factor = get_window_factor(data_settings["window"])

    extrinsic_prior_dict = get_extrinsic_prior_dict(data_settings["extrinsic_prior"])
    if data_settings["inference_parameters"] == "default":
        data_settings["inference_parameters"] = default_inference_parameters

    ref_time = data_settings["ref_time"]
    # Build detector objects
    ifo_list = InterferometerList(data_settings["detectors"])

    # Build transforms.
    transforms = [SampleExtrinsicParameters(extrinsic_prior_dict),
                  GetDetectorTimes(ifo_list, ref_time)]

    extra_context_parameters = []
    if "gnpe_time_shifts" in data_settings:
        d = data_settings["gnpe_time_shifts"]
        transforms.append(
            GNPECoalescenceTimes(
                ifo_list,
                d["kernel"],
                d["exact_equiv"],
                inference=False,
            )
        )
        extra_context_parameters += transforms[-1].context_parameters

    # Add the GNPE context to context_parameters the first time the transforms are
    # constructed. We do not want to overwrite the ordering of the parameters in
    # subsequent runs.
    if "context_parameters" not in data_settings:
        data_settings["context_parameters"] = []
    for p in extra_context_parameters:
        if p not in data_settings["context_parameters"]:
            data_settings["context_parameters"].append(p)

    # If the standardization factors have already been set, use those. Otherwise,
    # calculate them, and save them within the data settings.
    #
    # Standardizations are calculated at this point because the present set of
    # transforms is sufficient for generating samples of all regression and context
    # parameters.
    try:
        standardization_dict = data_settings["standardization"]
        print("Using previously-calculated parameter standardizations.")
    except KeyError:
        print("Calculating new parameter standardizations.")
        standardization_dict = get_standardization_dict(
            extrinsic_prior_dict,
            wfd,
            data_settings["inference_parameters"] + data_settings["context_parameters"],
            torchvision.transforms.Compose(transforms),
        )
        data_settings["standardization"] = standardization_dict

    transforms.append(ProjectOntoDetectors(ifo_list, domain, ref_time))
    transforms.append(SampleNoiseASD(asd_dataset))
    transforms.append(WhitenAndScaleStrain(domain.noise_std))
    # We typically add white detector noise. For debugging purposes, this can be turned
    # off with zero_noise option in data_settings.
    if not data_settings.get("zero_noise", False):
        transforms.append(AddWhiteNoiseComplex())
    transforms.append(
        SelectStandardizeRepackageParameters(
            {
                k: data_settings[k]
                for k in ["inference_parameters", "context_parameters"]
            },
            standardization_dict,
        )
    )
    transforms.append(
        RepackageStrainsAndASDS(data_settings["detectors"], first_index=domain.min_idx)
    )
    if data_settings["context_parameters"]:
        selected_keys = ["inference_parameters", "waveform", "context_parameters"]
    else:
        selected_keys = ["inference_parameters", "waveform"]

    transforms.append(UnpackDict(selected_keys=selected_keys))

    # Drop transforms that are not desired. This is useful for generating, e.g.,
    # noise-free data, or for producing data not formatted for input to the network.
    if omit_transforms is not None:
        transforms = [t for t in transforms if type(t) not in omit_transforms]

    wfd.transform = torchvision.transforms.Compose(transforms)


def build_svd_for_embedding_network(
    wfd: WaveformDataset,
    data_settings: dict,
    asd_dataset_path: str,
    size: int,
    num_training_samples: int,
    num_validation_samples: int,
    num_workers: int = 0,
    batch_size: int = 1000,
    out_dir=None,
):
    """
    Construct SVD matrices V based on clean waveforms in each interferometer. These
    will be used to seed the weights of the initial projection part of the embedding
    network.

    It first generates a number of training waveforms, and then produces the SVD.

    Parameters
    ----------
    wfd : WaveformDataset
    data_settings : dict
    asd_dataset_path : str
        Training waveforms will be whitened with respect to these ASDs.
    size : int
        Number of basis elements to include in the SVD projection.
    num_training_samples : int
    num_validation_samples : int
    num_workers : int
    batch_size : int
    out_dir : str
        SVD performance diagnostics are saved here.

    Returns
    -------
    list of numpy arrays
        The V matrices for each interferometer. They are ordered as in data_settings[
        'detectors'].
    """
    # Building the transforms can alter the data_settings dictionary. We do not want
    # the construction of the SVD to impact this, so begin with a fresh copy of this
    # dictionary.
    data_settings = copy.deepcopy(data_settings)

    # This is needed to prevent an occasional error when loading a large dataset into
    # memory using a dataloader. This removes a limitation on the number of "open files".
    old_sharing_strategy = torch.multiprocessing.get_sharing_strategy()
    torch.multiprocessing.set_sharing_strategy("file_system")

    # Fix the luminosity distance to a standard value, just in order to generate the SVD.
    data_settings["extrinsic_prior"]["luminosity_distance"] = "100.0"

    # Build the dataset, but with certain transforms omitted. In particular, we want to
    # build the SVD based on zero-noise waveforms. They should still be whitened though.
    set_train_transforms(
        wfd,
        data_settings,
        asd_dataset_path,
        omit_transforms=[
            AddWhiteNoiseComplex,
            RepackageStrainsAndASDS,
            SelectStandardizeRepackageParameters,
            UnpackDict,
        ],
    )

    print("Generating waveforms for embedding network SVD initialization.")
    time_start = time.time()
    ifos = list(wfd[0]["waveform"].keys())
    waveform_len = len(wfd[0]["waveform"][ifos[0]])
    num_waveforms = num_training_samples + num_validation_samples
    if num_waveforms > len(wfd):
        raise IndexError(
            f"Requested {num_waveforms} samples for generating SVD for embedding "
            f"network, but waveform dataset only contains {len(wfd)} samples."
        )
    waveforms = {
        ifo: np.empty((num_waveforms, waveform_len), dtype=np.complex128)
        for ifo in ifos
    }
    parameters = pd.DataFrame()

    loader = DataLoader(
        wfd,
        batch_size=batch_size,
        num_workers=num_workers,
        worker_init_fn=lambda _: np.random.seed(
            int(torch.initial_seed()) % (2 ** 32 - 1)
        ),
    )
    with threadpool_limits(limits=1, user_api="blas"):
        for idx, data in enumerate(loader):

            # This is for handling the last batch, which may otherwise push the total
            # number of samples above the number requested.
            lower = idx * batch_size
            n = min(batch_size, num_waveforms - lower)

            parameters = pd.concat(
                [parameters, pd.DataFrame(data["parameters"]).iloc[:n]],
                ignore_index=True,
            )
            strain_data = data["waveform"]
            for ifo, strains in strain_data.items():
                waveforms[ifo][lower : lower + n] = strains[:n]
            if lower + n == num_waveforms:
                break
    print(f"...done. This took {time.time() - time_start:.0f} s.")

    # Reset the standard sharing strategy.
    torch.multiprocessing.set_sharing_strategy(old_sharing_strategy)

    print("Generating SVD basis for ifo:")
    time_start = time.time()
    basis_dict = {}
    for ifo in ifos:
        basis = SVDBasis()
        basis.generate_basis(waveforms[ifo][:num_training_samples], size)
        basis_dict[ifo] = basis
        print(f"...{ifo} done.")
    print(f"...this took {time.time() - time_start:.0f} s.")

    if out_dir is not None:
        print(f"Testing SVD basis matrices.")
        for ifo, basis in basis_dict.items():
            print(f"...{ifo}:")
            basis.compute_test_mismatches(
                waveforms[ifo][num_training_samples:],
                parameters=parameters.iloc[num_training_samples:].reset_index(
                    drop=True
                ),
                verbose=True,
            )
            basis.to_file(os.path.join(out_dir, f"svd_{ifo}.hdf5"))
    print("Done")

    # Return V matrices in standard order. Drop the elements below domain.min_idx,
    # since the neural network expects data truncated below these. The dropped elements
    # should be 0.
    print(f"Truncating SVD matrices below index {wfd.domain.min_idx}.")
    print("...V matrix shapes:")
    V_rb_list = []
    for ifo in data_settings["detectors"]:
        V = basis_dict[ifo].V
        assert np.allclose(V[: wfd.domain.min_idx], 0)
        V = V[wfd.domain.min_idx :]
        print("      " + str(V.shape))
        V_rb_list.append(V)
    print("\n")
    return V_rb_list



import numpy as np
import torch
from bilby.gw.detector import PowerSpectralDensity
from scipy.interpolate import interp1d

from dingo.gw.domains import FrequencyDomain


class SampleNoiseASD(object):
    """
    Sample a random asds for each detector and add them to sample['asds'].
    """

    def __init__(self, asd_dataset):
        self.asd_dataset = asd_dataset

    def __call__(self, input_sample):
        sample = input_sample.copy()
        sample["asds"] = self.asd_dataset.sample_random_asds()
        return sample


class WhitenStrain(object):
    """
    Whiten the strain data by dividing w.r.t. the corresponding asds.
    """

    def __init__(self):
        pass

    def __call__(self, input_sample):
        sample = input_sample.copy()
        ifos = sample["waveform"].keys()
        if ifos != sample["asds"].keys():
            raise ValueError(
                f"Detectors of strain data, {ifos}, do not match "
                f'those of asds, {sample["asds"].keys()}.'
            )
        whitened_strains = {
            ifo: sample["waveform"][ifo] / sample["asds"][ifo] for ifo in ifos
        }
        sample["waveform"] = whitened_strains
        return sample


class WhitenFixedASD(object):
    """
    Whiten frequency-series data according to an ASD specified in a file. This uses the
    ASD files contained in Bilby.
    """

    def __init__(
        self,
        domain: FrequencyDomain,
        asd_file: str = None,
        inverse: bool = False,
        precision=None,
    ):
        """
        Parameters
        ----------
        domain : FrequencyDomain
            ASD is interpolated to the associated frequency grid.
        asd_file : str
            Name of the ASD file. If None, use the aligo ASD.
            [Default: None]
        inverse : bool
            Whether to apply the inverse whitening transform, to un-whiten data.
            [Default: False]
        precision : str ("single", "double")
            If not None, sets precision of ASD to specified precision.

        """
        if asd_file is not None:
            psd = PowerSpectralDensity(asd_file=asd_file)
        else:
            psd = PowerSpectralDensity.from_aligo()

        if psd.frequency_array[-1] < domain.f_max:
            raise ValueError(
                f"ASD in {asd_file} has f_max={psd.frequency_array[-1]}, "
                f"which is lower than domain f_max={domain.f_max}."
            )
        asd_interp = interp1d(
            psd.frequency_array, psd.asd_array, bounds_error=False, fill_value=np.inf
        )
        self.asd_array = asd_interp(domain.sample_frequencies)
        self.asd_array = domain.update_data(self.asd_array, low_value=1e-22)

        if precision is not None:
            if precision == "single":
                self.asd_array = self.asd_array.astype(np.float32)
            elif precision == "double":
                self.asd_array = self.asd_array.astype(np.float64)
            else:
                raise TypeError(
                    'precision can only be changed to "single" or "double".'
                )

        self.inverse = inverse

    def __call__(self, sample):
        """
        Parameters
        ----------
        sample : dict
            Dictionary of numpy arrays, e.g., with keys corresponding to polarizations.
            Method whitens each array with the same ASD.

        Returns
        -------
        dict of the same form as sample, but with whitened / un-whitened data.
        """
        result = {}
        for k, v in sample.items():
            if self.inverse:
                result[k] = v * self.asd_array
            else:
                result[k] = v / self.asd_array
        return result


class WhitenAndScaleStrain(object):
    """
    Whiten the strain data by dividing w.r.t. the corresponding asds,
    and scale it with 1/scale_factor.

    In uniform frequency domain the scale factor should be
    np.sqrt(window_factor) / np.sqrt(4.0 * delta_f).
    It has two purposes:
        (*) the denominator accounts for frequency binning
        (*) dividing by window factor accounts for windowing of strain data
    """

    def __init__(self, scale_factor):
        self.scale_factor = scale_factor

    def __call__(self, input_sample):
        sample = input_sample.copy()
        ifos = sample["waveform"].keys()
        if ifos != sample["asds"].keys():
            raise ValueError(
                f"Detectors of strain data, {ifos}, do not match "
                f'those of asds, {sample["asds"].keys()}.'
            )
        whitened_strains = {
            ifo: sample["waveform"][ifo] / (sample["asds"][ifo] * self.scale_factor)
            for ifo in ifos
        }
        sample["waveform"] = whitened_strains
        return sample


class AddWhiteNoiseComplex(object):
    """
    Adds white noise with a standard deviation determined by self.scale to the
    complex strain data.
    """

    def __init__(self):
        pass

    def __call__(self, input_sample):
        sample = input_sample.copy()
        noisy_strains = {}
        for ifo, pure_strain in sample["waveform"].items():
            # Use torch rng and convert to numpy, which is slightly faster than using
            # numpy directly. Using torch.randn gives single-precision floats by default
            # (which we want)  whereas np.random.random gives double precision (and
            # must subsequently  be cast to single precision).
            # np.random.default_rng().standard_normal() can be set to output single
            # precision, but in testing this is slightly slower than the torch call.
            noise = (
                torch.randn(len(pure_strain), device=torch.device("cpu"))
                + torch.randn(len(pure_strain), device=torch.device("cpu")) * 1j
            )
            noise = noise.numpy()
            noisy_strains[ifo] = pure_strain + noise
        sample["waveform"] = noisy_strains
        return sample


class RepackageStrainsAndASDS(object):
    """
    Repackage the strains and the asds into an [num_ifos, 3, num_bins]
    dimensional tensor. Order of ifos is provided by self.ifos. By
    convention, [:,i,:] is used for:
        i = 0: strain.real
        i = 1: strain.imag
        i = 2: 1 / (asd * 1e23)
    """

    def __init__(self, ifos, first_index=0):
        self.ifos = ifos
        self.first_index = first_index

    def __call__(self, input_sample):
        sample = input_sample.copy()
        strains = np.empty(
            (len(self.ifos), 3, len(sample["asds"][self.ifos[0]]) - self.first_index),
            dtype=np.float32,
        )
        for idx_ifo, ifo in enumerate(self.ifos):
            strains[idx_ifo, 0] = sample["waveform"][ifo][self.first_index :].real
            strains[idx_ifo, 1] = sample["waveform"][ifo][self.first_index :].imag
            strains[idx_ifo, 2] = 1 / (sample["asds"][ifo][self.first_index :] * 1e23)
        sample["waveform"] = strains
        return sample


# if __name__ == "__main__":
#     AD = ASDDataset("../../../data/PSDs/asds_O1.hdf5")
#     asd_samples = AD.sample_random_asds()



import numpy as np
import pycbc.psd
from gwpy.timeseries import TimeSeries

from dingo.gw.domains import FrequencyDomain
from dingo.gw.gwutils import (
    get_window,
    get_window_factor,
)


def estimate_single_psd(
    time_start,
    time_segment,
    window,
    f_s=4096,
    num_segments: int = 128,
    det=None,
    channel=None,
):
    """
    Download strain data and generate a PSD based on these. Use num_segments of length
    time_segment, starting at GPS time time_start. If no channel is specified, GWOSC is used
    to download the data.

    Parameters
    ----------
    time_start: float
        start GPS time for PSD estimation
    time_segment: float
        time for a single segment for PSD information, in seconds
    window: Union(np.ndarray, dict)
        Window used for PSD generation, needs to be the same as used for Fourier
        transform of event strain data.
        Either provided directly as np.ndarray, or as dict in which case the window is
        generated by window = dingo.gw.gwutils.get_window(**window).
    num_segments: int = 256
        number of segments used for PSD generation
    det: str
        If provided, will download data from GWOSC using TimeSeries.fetch_open_data(). Mutually exclusive with 'channel'.
    channel: str
        If provided, will download the data from the channel instead of gwosc using TimeSeries.get()
    Returns
    -------
    psd: np.array
        array of psd
    """
    # download strain data for psd
    time_end = time_start + time_segment * num_segments
    assert (
        det is not None and channel is None or det is None and channel is not None
    ), "det and channel are mutually exclusive!"

    if channel:
        psd_strain = TimeSeries.get(channel, time_start, time_end)
        # TODO: We currently assume that sample rate of channel matches that provided in the settings?
    else:
        psd_strain = TimeSeries.fetch_open_data(
            det, time_start, time_end, sample_rate=f_s, cache=False
        )
    psd_strain = psd_strain.to_pycbc()

    # optionally generate window
    if type(window) == dict:
        window = get_window(**window)
    assert (
        len(window) == len(psd_strain) / num_segments
    ), "Window does not match strain. Is sampling frequency f_s off?"

    # generate PSD from strain data
    psd = pycbc.psd.estimate.welch(
        psd_strain,
        seg_len=len(window),
        seg_stride=len(window),
        window=window,
        avg_method="median",
    )

    return np.array(psd)


def download_strain_data_in_FD(det, time_event, time_segment, time_buffer, window):
    """
    Download strain data for a GW event at GPS time time_event. The segment is
    time_segment seconds long, including time_buffer seconds after the event. The
    strain is Fourier transformed, the frequency domain strain is then time shifted by
    time_buffer, such that the event occurs at t=0.

    Parameters
    ----------
    det: str
        detector
    time_event: float
        GPS time of the event
    time_segment: float
        length of the strain segment, in seconds
    time_buffer: float
        specifies buffer time after time_event, in seconds
    window: Union(np.ndarray, dict)
        Window used for Fourier transforming the event strain data.
        Either provided directly as np.ndarray, or as dict in which case the window is
        generated by window = dingo.gw.gwutils.get_window(**window).


    Returns
    -------
    event_strain: np.array
        array with the frequency domain strain
    """
    # download strain data
    print("Downloading strain data for event.", end=" ")
    event_strain = TimeSeries.fetch_open_data(
        det,
        time_event + time_buffer - time_segment,
        time_event + time_buffer,
        cache=True,
    )
    print("Done.")

    # transform to FD
    if type(window) == dict:
        window = get_window(**window)
    assert len(window) == len(
        event_strain
    ), "Window does not match strain. Is sampling frequency f_s off?"
    event_strain = event_strain.to_pycbc()
    event_strain = (event_strain * window).to_frequencyseries()

    # time shift by time_buffer, such that event happens at time 0
    event_strain = event_strain.cyclic_time_shift(time_buffer)

    return np.array(event_strain)


def download_event_data_in_FD(
    detectors,
    time_event,
    time_segment,
    time_buffer,
    window,
    num_segments_psd=128,
):
    """
    Download event data in frequency domain. This includes strain data for the event at
    GPS time t_event as well as the correcponding ASD.

    Parameters
    ----------
    detectors: list
        list of detectors specified via strings
    time_event: float
        GPS time of the event
    time_segment: float
        length of the strain segment, in seconds
    time_buffer: float
        specifies buffer time after time_event, in seconds
    window: Union(np.ndarray, dict)
        Window used for Fourier transforming the event strain data.
        Either provided directly as np.ndarray, or as dict in which case the window is
        generated by window = dingo.gw.gwutils.get_window(**window).
    num_segments_psd: int = 128
        number of segments used for PSD generation

    Returns
    -------
    """
    data = {"waveform": {}, "asds": {}}
    for det in detectors:
        print("Detector {:}:".format(det))

        data["waveform"][det] = download_strain_data_in_FD(
            det, time_event, time_segment, time_buffer, window
        )
        data["asds"][det] = (
            estimate_single_psd(
                time_event + time_buffer - time_segment * (num_segments_psd + 1),
                time_segment,
                window,
                num_segments_psd,
                det=det
            )
            ** 0.5
        )

    # build domain object
    f_s = len(data["waveform"][detectors[0]]) / time_segment
    domain = FrequencyDomain(
        f_min=0,
        f_max=f_s / 2,
        delta_f=1 / time_segment,
        window_factor=get_window_factor(window),
    )

    return data, domain



from itertools import zip_longest

import matplotlib as mpl
import matplotlib.pyplot as plt
import corner
import numpy as np
import pandas as pd


def plot_corner_multi(
    samples, weights=None, labels=None, filename="corner.pdf", **kwargs
):
    """
    Generate a corner plot for multiple posteriors.

    Parameters
    ----------
    samples : list[pd.DataFrame]
        List of sample sets. The DataFrame column names are used as parameter labels.
    weights : list[np.ndarray or None] or None
        List of weights sets. The length of each array should be the same as the length of
        the corresponding samples.
    labels : list[str or None] or None
        Labels for the posteriors.
    filename : str
        Where to save samples.
    **kwargs :
        Forwarded to corner.corner.
    """
    # Define plot properties
    cmap = "Dark2"
    corner_params = {
        "smooth": 1.0,
        "smooth1d": 1.0,
        "plot_datapoints": False,
        "plot_density": False,
        "plot_contours": True,
        "levels": [0.5, 0.9],
        "bins": 30,
    }
    corner_params.update(kwargs)

    serif_old = mpl.rcParams["font.family"]
    mpl.rcParams["font.family"] = "serif"
    linewidth_old = mpl.rcParams["lines.linewidth"]
    mpl.rcParams["lines.linewidth"] = 2.5

    # In case a single corner plot is desired, convert to lists to iterate.
    if not isinstance(samples, list):
        samples = [samples]
    if not isinstance(weights, list):
        weights = [weights]
    if not isinstance(labels, list):
        labels = [labels]

    # Only plot common parameters for all sample sets, keeping the same order as the
    # first sample set.
    common_parameters = [
        p
        for p in samples[0].columns
        if p in set.intersection(*(set(s.columns) for s in samples))
    ]

    fig = None
    handles = []
    for i, (s, w, l) in enumerate(zip_longest(samples, weights, labels)):
        color = mpl.colors.rgb2hex(plt.get_cmap(cmap)(i))
        fig = corner.corner(
            s[common_parameters].to_numpy(),
            labels=common_parameters,
            weights=w,
            color=color,
            no_fill_contours=True,
            fig=fig,
            **corner_params
        )
        handles.append(
            plt.Line2D([], [], color=color, label=l, linewidth=5, markersize=20)
        )

    # Eliminate spacing between the 2D plots
    if len(common_parameters) > 8:
        fig.subplots_adjust(wspace=0, hspace=0)
    else:
        space = 1 / (4 * len(common_parameters))
        fig.subplots_adjust(wspace=space, hspace=space)

    fig.legend(
        handles=handles,
        loc="upper right",
        fontsize=50,
        labelcolor="linecolor",
    )

    # Customize tick and label properties for each axis
    for i, ax in enumerate(fig.get_axes()):
        if ax.get_xlabel():
            ax.tick_params(
                axis="x", labelsize=14, length=6, width=1.5
            )  # Adjust labelsize, length, and width
            ax.xaxis.label.set_size(16)  # Adjust x-axis label font size
        else:
            ax.tick_params(axis="x", which="both", bottom=False)
        if ax.get_ylabel():
            ax.tick_params(
                axis="y", labelsize=14, length=6, width=1.5
            )  # Adjust labelsize, length, and width
            ax.yaxis.label.set_size(16)  # Adjust x-axis label font size
        else:
            ax.tick_params(axis="y", which="both", left=False)

        # Turn off grid
        ax.grid()

    # Save the figure
    plt.savefig(filename)

    # Reset rcParams to original values
    mpl.rcParams["font.family"] = serif_old
    mpl.rcParams["lines.linewidth"] = linewidth_old

    return fig



import torch
import argparse


def parse_args():

    parser = argparse.ArgumentParser()
    parser.add_argument('--checkpoint', type=str, required=True)
    parser.add_argument('--key', type=str, required=True, nargs='+')
    parser.add_argument('--value', type=str, required=True)
    return parser.parse_args()


def main():
    args = parse_args()

    # Typically training is done on the GPU, so the model could be saved on a GPU
    # device. Since this routine may be run on a CPU machine, allow for a remap of the
    # torch tensors.
    if torch.cuda.is_available():
        d = torch.load(args.checkpoint)
    else:
        d = torch.load(args.checkpoint, map_location=torch.device('cpu'))

    # Figure out type of value
    try:
        value = int(args.value)
    except ValueError:
        try:
            value = float(args.value)
        except ValueError:
            value = args.value

    settings = d['metadata']
    for i in args.key[:-1]:
        settings = settings[i]
    settings[args.key[-1]] = value

    torch.save(d, args.checkpoint)


if __name__ == "__main__":
    main()